# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Optional

import jsonschema

from ._api_schema import RELEASES_JSONSCHEMA

"""
This module holds representations for results for the v2 releases
API endpoint provided by the Snap Store. It currently only covers
Revisions and Releases.

The full API is documented on
https://dashboard.snapcraft.io/docs/v2/en/snaps.html#snap-releases
"""


class Release:
    """Represent a snap release."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Release":
        jsonschema.validate(
            payload,
            RELEASES_JSONSCHEMA["properties"]["releases"]["items"]["properties"],
        )
        return cls(
            architecture=payload["architecture"],
            branch=payload["branch"],
            channel=payload["channel"],
            expiration_date=payload["expiration-date"],
            revision=payload["revision"],
            risk=payload["risk"],
            track=payload["track"],
            when=payload["when"],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "architecture": self.architecture,
            "branch": self.branch,
            "channel": self.channel,
            "expiration-date": self.expiration_date,
            "revision": self.revision,
            "risk": self.risk,
            "track": self.track,
            "when": self.when,
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.channel!r}>"

    def __init__(
        self,
        architecture: str,
        branch: Optional[str],
        channel: str,
        expiration_date: Optional[str],
        revision: Optional[int],
        risk: str,
        track: str,
        when: str,
    ) -> None:
        self.architecture = architecture
        self.branch = branch
        self.channel = channel
        self.expiration_date = expiration_date
        self.revision = revision
        self.risk = risk
        self.track = track
        self.when = when


class Revision:
    """Represent a snap revision."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Revision":
        jsonschema.validate(
            payload,
            RELEASES_JSONSCHEMA["properties"]["revisions"]["items"]["properties"],
        )
        return cls(
            architectures=payload["architectures"],
            base=payload.get("base"),
            build_url=payload["build_url"],
            confinement=payload["confinement"],
            created_at=payload["created_at"],
            grade=payload["grade"],
            revision=payload["revision"],
            sha3_384=payload["sha3-384"],
            size=payload["size"],
            status=payload["status"],
            version=payload["version"],
        )

    def marshal(self) -> Dict[str, Any]:
        payload = {
            "architectures": self.architectures,
            "build_url": self.build_url,
            "confinement": self.confinement,
            "created_at": self.created_at,
            "grade": self.grade,
            "revision": self.revision,
            "sha3-384": self.sha3_384,
            "size": self.size,
            "status": self.status,
            "version": self.version,
        }

        if self.base is not None:
            payload["base"] = self.base

        return payload

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.revision!r}>"

    def __init__(
        self,
        *,
        architectures: List[str],
        base: Optional[str],
        build_url: Optional[str],
        confinement: str,
        created_at: str,
        grade: str,
        revision: int,
        sha3_384: str,
        size: int,
        status: str,
        version: str,
    ):
        self.architectures = architectures
        self.base = base
        self.build_url = build_url
        self.confinement = confinement
        self.created_at = created_at
        self.grade = grade
        self.revision = revision
        self.sha3_384 = sha3_384
        self.size = size
        self.status = status
        self.version = version


class Releases:
    """Represent the data returned from the releases Snap Store endpoint."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Releases":
        jsonschema.validate(payload, RELEASES_JSONSCHEMA)
        return cls(
            releases=[Release.unmarshal(r) for r in payload["releases"]],
            revisions=[Revision.unmarshal(r) for r in payload["revisions"]],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "releases": [r.marshal() for r in self.releases],
            "revisions": [r.marshal() for r in self.revisions],
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def __init__(self, releases: List[Release], revisions: List[Revision]) -> None:
        self.releases = releases
        self.revisions = revisions



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
from unittest import mock

import fixtures
from testtools.matchers import Equals

from snapcraft.internal import sources
from tests import unit


# LP: #1733584
class TestSubversion(unit.sources.SourceTestCase):  # type: ignore
    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.sources.Subversion._get_source_details")
        self.mock_get_source_details = patcher.start()
        self.mock_get_source_details.return_value = ""
        self.addCleanup(patcher.stop)

    def test_pull_remote(self):
        svn = sources.Subversion("svn://my-source", "source_dir")
        svn.pull()
        self.mock_run.assert_called_once_with(
            ["svn", "checkout", "svn://my-source", "source_dir"]
        )

    def test_pull_remote_commit(self):
        svn = sources.Subversion("svn://my-source", "source_dir", source_commit="2")
        svn.pull()
        self.mock_run.assert_called_once_with(
            ["svn", "checkout", "svn://my-source", "source_dir", "-r", "2"]
        )

    def test_pull_local_absolute_path(self):
        svn = sources.Subversion(self.path, "source_dir")
        svn.pull()
        self.mock_run.assert_called_once_with(
            ["svn", "checkout", "file://" + self.path, "source_dir"]
        )

    def test_pull_local_relative_path(self):
        os.mkdir("my-source")
        svn = sources.Subversion("my-source", "source_dir")
        svn.pull()
        self.mock_run.assert_called_once_with(
            [
                "svn",
                "checkout",
                "file://{}".format(os.path.join(self.path, "my-source")),
                "source_dir",
            ]
        )

    def test_pull_existing(self):
        self.mock_path_exists.return_value = True
        svn = sources.Subversion("svn://my-source", "source_dir")
        svn.pull()
        self.mock_run.assert_called_once_with(["svn", "update"], cwd=svn.source_dir)

    def test_init_with_source_tag_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceInvalidOptionError,
            sources.Subversion,
            "svn://mysource",
            "source_dir",
            source_tag="tag",
        )
        self.assertThat(raised.source_type, Equals("svn"))
        self.assertThat(raised.option, Equals("source-tag"))

    def test_init_with_source_branch_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceInvalidOptionError,
            sources.Subversion,
            "svn://mysource",
            "source_dir",
            source_branch="branch",
        )
        self.assertThat(raised.source_type, Equals("svn"))
        self.assertThat(raised.option, Equals("source-branch"))

    def test_init_with_source_depth_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceInvalidOptionError,
            sources.Subversion,
            "svn://mysource",
            "source_dir",
            source_depth=2,
        )

        self.assertThat(raised.source_type, Equals("svn"))
        self.assertThat(raised.option, Equals("source-depth"))

    def test_source_checksum_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceInvalidOptionError,
            sources.Subversion,
            "svn://mysource",
            "source_dir",
            source_checksum="md5/d9210476aac5f367b14e513bdefdee08",
        )

        self.assertThat(raised.source_type, Equals("svn"))
        self.assertThat(raised.option, Equals("source-checksum"))

    def test_has_source_handler_entry(self):
        self.assertTrue(sources._source_handler["subversion"] is sources.Subversion)

    def test_pull_failure(self):
        self.mock_run.side_effect = subprocess.CalledProcessError(1, [])

        svn = sources.Subversion("svn://my-source", "source_dir")
        raised = self.assertRaises(sources.errors.SnapcraftPullError, svn.pull)
        self.assertThat(
            raised.command, Equals("svn checkout svn://my-source source_dir")
        )
        self.assertThat(raised.exit_code, Equals(1))


def get_side_effect(original_call):
    def side_effect(cmd, *args, **kwargs):
        if len(cmd) > 1 and cmd[1] == "info":
            return "mock-commit".encode()
        elif cmd[0] == "svn":
            return
        return original_call(cmd, *args, **kwargs)

    return side_effect


class SubversionDetailsTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.repo_tree = "svn-repo"
        self.working_tree = "svn-test"
        self.source_dir = "svn-checkout"

        self.expected_commit = "1"

        self.fake_check_output = self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_output",
                side_effect=get_side_effect(subprocess.check_output),
            )
        )
        self.fake_check_call = self.useFixture(
            fixtures.MockPatch(
                "subprocess.check_call",
                side_effect=get_side_effect(subprocess.check_call),
            )
        )

    def test_svn_details_commit(self):
        svn = sources.Subversion(self.repo_tree, self.source_dir, silent=True)
        svn.pull()

        source_details = svn._get_source_details()
        self.assertThat(source_details["source-commit"], Equals("mock-commit"))

        self.fake_check_call.mock.assert_called_once_with(
            ["svn", "checkout", "svn-repo", "svn-checkout"], stderr=-3, stdout=-3
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import contextlib
import functools
import glob
import logging
import os
import pathlib
import re
import shutil
import subprocess
import tempfile
from typing import Dict, FrozenSet, List, Optional, Sequence, Set, Tuple, Union

import elftools.common.exceptions
import elftools.elf.elffile
from elftools.construct import ConstructError
from pkg_resources import parse_version

from snapcraft import file_utils
from snapcraft.internal import common, errors, repo
from snapcraft.project._project_options import ProjectOptions

logger = logging.getLogger(__name__)


@functools.lru_cache()
def _get_host_libc_path() -> pathlib.Path:
    arch_triplet = ProjectOptions().arch_triplet
    return pathlib.Path("/lib") / arch_triplet / "libc.so.6"


def _ldd_resolve(soname: str, soname_path: str) -> Tuple[str, str]:
    logger.debug(f"_ldd_resolve: {soname!r} {soname_path!r}")

    # If found, resolve the path components.  We can safely determine that
    # ldd found the match if it returns an absolute path.  For additional
    # safety, check that it exists.  See example ldd output in ldd() below.
    # If not found, ldd should use a string like "not found", but we do not
    # really care what that string is with this approach as it has to start
    # with "/" and point to a valid file.
    if soname_path.startswith("/") and os.path.exists(soname_path):
        abs_path = os.path.abspath(soname_path)
        return soname, abs_path

    # Not found, use the soname.
    return soname, soname


def _check_output(cmd: List[str], *, extra_env: Dict[str, str]) -> str:
    env = os.environ.copy()
    env.update(extra_env)

    debug_cmd = [f"{k}={v}" for k, v in extra_env.items()]
    debug_cmd += cmd

    logger.debug("executing: %s", " ".join(debug_cmd))
    output = subprocess.check_output(cmd, env=env).decode()

    return output


def _parse_ldd_output(output: str) -> Dict[str, str]:
    """Parse ldd output.

    Example ldd outputs:

    linux-vdso.so.1 =>  (0x00007ffdc13ec000)   <== ubuntu 16.04 ldd
    linux-vdso.so.1 (0x00007ffdc13ec000)       <== newer ldd
    /lib64/ld-linux-x86-64.so.2 (0x00007fb3c5298000)
    libm.so.6 => /lib/x86_64-linux-gnu/libm.so.6 (0x00007fb3bef03000)
    libmissing.so.2 => not found

    :returns: Dictionary of dependencies, mapping library name to path.
    """
    libraries: Dict[str, str] = {}
    ldd_lines = output.splitlines()

    for line in ldd_lines:
        # First match against libraries that are found.
        match = re.match(r"\t(.*) => (.*) \(0x", line)

        if not match:
            # Now find those not found, or not providing the address...
            match = re.match(r"\t(.*) => (.*)", line)

        # Ignore ld-linux, linux-vdso, etc. that don't match these regex.
        # As Ubuntu 16.04's ldd provides an empty string for the found
        # path (in group 2) on linux-vdso, check for this and ignore it.
        # See example output above for reference.
        if not match or match.group(2) == "":
            continue

        soname, soname_path = _ldd_resolve(match.group(1), match.group(2))
        libraries[soname] = soname_path

    return libraries


def _ld_trace(path: str, ld_library_paths: List[str]) -> Dict[str, str]:
    """Use LD_TRACE_LOADED_OBJECTS to determine library dependencies."""
    env = {
        "LD_TRACE_LOADED_OBJECTS": "1",
        "LD_LIBRARY_PATH": ":".join(ld_library_paths),
    }

    return _parse_ldd_output(_check_output([path], extra_env=env))


def _ldd(
    path: str, ld_library_paths: List[str], *, ld_preload: Optional[str] = None
) -> Dict[str, str]:
    """Use host ldd to determine library dependencies."""
    ldd_path = str(
        file_utils.get_host_tool_path(command_name="ldd", package_name="libc-bin")
    )
    env = {
        "LD_LIBRARY_PATH": ":".join(ld_library_paths),
    }

    if ld_preload:
        env["LD_PRELOAD"] = ld_preload

    return _parse_ldd_output(_check_output([ldd_path, path], extra_env=env))


def _determine_libraries(*, path: str, ld_library_paths: List[str]) -> Dict[str, str]:
    # Try the usual method with ldd.
    with contextlib.suppress(subprocess.CalledProcessError):
        return _ldd(path, ld_library_paths)

    # Fall back to trying ldd with LD_PRELOAD explicitly loading libc.
    libc_path = _get_host_libc_path()
    if libc_path.is_file():
        with contextlib.suppress(subprocess.CalledProcessError):
            return _ldd(path, ld_library_paths, ld_preload=str(libc_path))

    # Fall back to trying ld trace method which may fail with permission error
    # for non-executable shared objects, or OSError 8 Exec format error if
    # target is for different arch.
    with contextlib.suppress(PermissionError, OSError, subprocess.CalledProcessError):
        return _ld_trace(path, ld_library_paths)

    logger.warning("Unable to determine library dependencies for %r", path)
    return {}


class NeededLibrary:
    """Represents an ELF library version."""

    def __init__(self, *, name: str) -> None:
        self.name = name
        self.versions = set()  # type: Set[str]

    def add_version(self, version: str) -> None:
        self.versions.add(version)


ElfArchitectureTuple = Tuple[str, str, str]
SonameCacheDict = Dict[Tuple[ElfArchitectureTuple, str], str]


# Old pyelftools uses byte strings for section names.  Some data is
# also returned as bytes, which is handled below.
if parse_version(elftools.__version__) >= parse_version("0.24"):
    _DEBUG_INFO: Union[str, bytes] = ".debug_info"
    _DYNAMIC: Union[str, bytes] = ".dynamic"
    _GNU_VERSION_D: Union[str, bytes] = ".gnu.version_d"
    _GNU_VERSION_R: Union[str, bytes] = ".gnu.version_r"
    _INTERP: Union[str, bytes] = ".interp"
else:
    _DEBUG_INFO = b".debug_info"
    _GNU_VERSION_D = b".gnu.version_d"
    _GNU_VERSION_R = b".gnu.version_r"


class SonameCache:
    """A cache for sonames."""

    def __getitem__(self, key):
        return self._soname_paths[key]

    def __setitem__(self, key, item):
        # Initial API error checks
        if not isinstance(key, tuple):
            raise EnvironmentError(
                "The key for SonameCache has to be a (arch, soname) tuple."
            )
        if not isinstance(key[0], tuple) or len(key[0]) != 3:
            raise EnvironmentError(
                "The first element of the key needs to of type ElfArchitectureTuple."
            )
        if not isinstance(key[1], str):
            raise EnvironmentError(
                "The second element of the key needs to be "
                "of type str representing the soname."
            )
        self._soname_paths[key] = item

    def __contains__(self, key):
        return key in self._soname_paths

    def __init__(self):
        """Initialize a cache for sonames"""
        self._soname_paths = dict()  # type: SonameCacheDict

    def reset_except_root(self, root):
        """Reset the cache values that aren't contained within root."""
        new_soname_paths = dict()  # type: SonameCacheDict
        for key, value in self._soname_paths.items():
            if value is not None and value.startswith(root):
                new_soname_paths[key] = value

        self._soname_paths = new_soname_paths


class Library:
    """Represents the SONAME and path to the library."""

    def __init__(
        self,
        *,
        soname: str,
        soname_path: str,
        search_paths: List[str],
        core_base_path: Optional[str],
        arch: ElfArchitectureTuple,
        soname_cache: SonameCache,
    ) -> None:

        self.soname = soname
        self.soname_path = soname_path
        self.search_paths = search_paths
        self.core_base_path = core_base_path
        self.arch = arch
        self.soname_cache = soname_cache

        # Resolve path, if possible.
        self.path = self._crawl_for_path()

        if core_base_path is not None and self.path.startswith(core_base_path):
            self.in_base_snap = True
        else:
            self.in_base_snap = False

        logger.debug(
            "{soname} with original path {original_path} found on {path} in base: {in_base}".format(
                soname=soname,
                original_path=soname_path,
                path=self.path,
                in_base=self.in_base_snap,
            )
        )

    def _update_soname_cache(self, resolved_path: str) -> None:
        self.soname_cache[self.arch, self.soname] = resolved_path

    def _is_valid_elf(self, resolved_path: str) -> bool:
        if not os.path.exists(resolved_path) or not ElfFile.is_elf(resolved_path):
            return False

        try:
            elf_file = ElfFile(path=resolved_path)
        except errors.CorruptedElfFileError as error:
            # Log if the ELF file seems corrupted.
            logger.warning(error.get_brief())
            return False

        return elf_file.arch == self.arch

    def _crawl_for_path(self) -> str:
        # Speed things up and return what was already found once.
        if (self.arch, self.soname) in self.soname_cache:
            return self.soname_cache[self.arch, self.soname]

        logger.debug("Crawling to find soname {!r}".format(self.soname))

        valid_search_paths = [p for p in self.search_paths if os.path.exists(p)]
        in_search_paths = any(
            self.soname_path.startswith(p) for p in valid_search_paths
        )

        # Expedite path crawling if we have a valid elf file that lives
        # inside the search paths.
        if in_search_paths and self._is_valid_elf(self.soname_path):
            self._update_soname_cache(self.soname_path)
            return self.soname_path

        for path in valid_search_paths:
            for root, directories, files in os.walk(path):
                if self.soname not in files:
                    continue

                file_path = os.path.join(root, self.soname.strip("/"))
                if self._is_valid_elf(file_path):
                    self._update_soname_cache(file_path)
                    return file_path

        # Required for libraries on the host and the fetching mechanism.
        self._update_soname_cache(self.soname_path)
        return self.soname_path


# Old versions of pyelftools return bytes rather than strings for
# certain APIs.  So we pass those values through this function to get
# a consistent result.
def _ensure_str(s):
    if isinstance(s, bytes):
        return s.decode("ascii")
    assert isinstance(s, str)
    return s


class ElfFile:
    """ElfFile represents and elf file on a path and its attributes."""

    @classmethod
    def is_elf(cls, path: str) -> bool:
        if not os.path.isfile(path):
            # ELF binaries are regular files
            return False
        with open(path, "rb") as bin_file:
            return bin_file.read(4) == b"\x7fELF"

    def __init__(self, *, path: str) -> None:
        """Initialize an ElfFile instance.

        :param str path: path to an elf_file within a snapcraft project.
        """
        self.path = path
        self.dependencies = set()  # type: Set[Library]

        self.arch: Optional[ElfArchitectureTuple] = None
        self.interp: str = ""
        self.soname: str = ""
        self.versions: Set[str] = set()
        self.needed: Dict[str, NeededLibrary] = dict()
        self.execstack_set: bool = False
        self.is_dynamic: bool = True
        self.build_id: str = ""
        self.has_debug_info: bool = False

        # String of elf enum type, e.g. "ET_DYN", "ET_EXEC", etc.
        self.elf_type: str = "ET_NONE"

        try:
            logger.debug(f"Extracting ELF attributes: {path}")
            self._extract_attributes()
        except (UnicodeDecodeError, AttributeError, ConstructError) as exception:
            logger.debug(f"Extracting ELF attributes exception: {str(exception)}")
            raise errors.CorruptedElfFileError(path, exception)

    def _extract_attributes(self) -> None:  # noqa: C901
        with open(self.path, "rb") as fp:
            elf = elftools.elf.elffile.ELFFile(fp)

            # A set of fields to identify the architecture of the ELF file:
            #  EI_CLASS: 32/64 bit (e.g. amd64 vs. x32)
            #  EI_DATA: byte orer (e.g. ppc64 vs. ppc64le)
            #  e_machine: instruction set (e.g. x86-64 vs. arm64)
            #
            # For amd64 binaries, this will evaluate to:
            #   ('ELFCLASS64', 'ELFDATA2LSB', 'EM_X86_64')
            self.arch = (
                elf.header.e_ident.EI_CLASS,
                elf.header.e_ident.EI_DATA,
                elf.header.e_machine,
            )

            # Gather attributes from dynamic sections.
            for section in elf.iter_sections():
                if not isinstance(section, elftools.elf.dynamic.DynamicSection):
                    continue

                self.is_dynamic = True

                for tag in section.iter_tags():
                    if tag.entry.d_tag == "DT_NEEDED":
                        needed = _ensure_str(tag.needed)
                        self.needed[needed] = NeededLibrary(name=needed)
                    elif tag.entry.d_tag == "DT_SONAME":
                        self.soname = _ensure_str(tag.soname)

            for segment in elf.iter_segments():
                if segment["p_type"] == "PT_GNU_STACK":
                    # p_flags holds the bit mask for this segment.
                    # See `man 5 elf`.
                    mode = segment["p_flags"]
                    if mode & elftools.elf.constants.P_FLAGS.PF_X:
                        self.execstack_set = True
                elif isinstance(segment, elftools.elf.segments.InterpSegment):
                    self.interp = segment.get_interp_name()

            build_id_section = elf.get_section_by_name(".note.gnu.build-id")
            if (
                isinstance(build_id_section, elftools.elf.sections.NoteSection)
                and build_id_section.header["sh_type"] != "SHT_NOBITS"
            ):
                for note in build_id_section.iter_notes():
                    if note.n_name == "GNU" and note.n_type == "NT_GNU_BUILD_ID":
                        self.build_id = _ensure_str(note.n_desc)

            # If we are processing a detached debug info file, these
            # sections will be present but empty.
            verneed_section = elf.get_section_by_name(_GNU_VERSION_R)
            if isinstance(verneed_section, elftools.elf.gnuversions.GNUVerNeedSection):
                for library, versions in verneed_section.iter_versions():
                    library_name = _ensure_str(library.name)
                    # If the ELF file only references weak symbols
                    # from a library, it may be absent from DT_NEEDED
                    # but still have an entry in .gnu.version_r for
                    # symbol versions.
                    if library_name not in self.needed:
                        continue
                    lib = self.needed[library_name]
                    for version in versions:
                        lib.add_version(_ensure_str(version.name))

            verdef_section = elf.get_section_by_name(_GNU_VERSION_D)
            if isinstance(verdef_section, elftools.elf.gnuversions.GNUVerDefSection):
                for _, auxiliaries in verdef_section.iter_versions():
                    for aux in auxiliaries:
                        self.versions.add(_ensure_str(aux.name))

            debug_info_section = elf.get_section_by_name(_DEBUG_INFO)
            self.has_debug_info = (
                debug_info_section is not None
                and debug_info_section.header.sh_type != "SHT_NOBITS"
            )

            self.elf_type = elf.header["e_type"]

    def is_linker_compatible(self, *, linker_version: str) -> bool:
        """Determines if linker will work given the required glibc version."""
        version_required = self.get_required_glibc()
        r = parse_version(version_required) <= parse_version(linker_version)
        logger.debug(
            "Checking if linker {!r} will work with "
            "GLIBC_{} required by {!r}: {!r}".format(
                linker_version, version_required, self.path, r
            )
        )
        return r

    def get_required_glibc(self) -> str:
        """Returns the required glibc version for this ELF file."""
        with contextlib.suppress(AttributeError):
            return self._required_glibc  # type: ignore

        version_required = ""
        for lib in self.needed.values():
            for version in lib.versions:
                if not version.startswith("GLIBC_"):
                    continue
                version = version[6:]
                if parse_version(version) > parse_version(version_required):
                    version_required = version

        self._required_glibc = version_required
        return version_required

    def load_dependencies(
        self,
        root_path: str,
        core_base_path: Optional[str],
        content_dirs: Set[str],
        arch_triplet: str,
        soname_cache: SonameCache = None,
    ) -> Set[str]:
        """Load the set of libraries that are needed to satisfy elf's runtime.

        This may include libraries contained within the project.
        The object's .dependencies attribute is set after loading.

        :param str root_path: the root path to search for missing dependencies.
        :param str core_base_path: the core base path to search for missing
                                   dependencies.
        :param SonameCache soname_cache: a cache of previously search
                                         dependencies.
        :returns: a set of string with paths to the library dependencies of
                  elf.
        """
        if soname_cache is None:
            soname_cache = SonameCache()

        logger.debug("Getting dependencies for {!r}".format(self.path))

        search_paths = [root_path, *content_dirs]
        if core_base_path is not None:
            search_paths.append(core_base_path)

        ld_library_paths: List[str] = list()
        for path in search_paths:
            ld_library_paths.extend(common.get_library_paths(path, arch_triplet))

        libraries = _determine_libraries(
            path=self.path, ld_library_paths=ld_library_paths
        )
        for soname, soname_path in libraries.items():
            if self.arch is None:
                raise RuntimeError("failed to parse architecture")

            self.dependencies.add(
                Library(
                    soname=soname,
                    soname_path=soname_path,
                    search_paths=search_paths,
                    core_base_path=core_base_path,
                    arch=self.arch,
                    soname_cache=soname_cache,
                )
            )

        # Return the set of dependency paths, minus those found in the base.
        dependencies: Set[str] = set()
        for library in self.dependencies:
            if not library.in_base_snap:
                dependencies.add(library.path)
        return dependencies


class Patcher:
    """Patcher holds the necessary logic to patch elf files."""

    def __init__(
        self, *, dynamic_linker: str, root_path: str, preferred_patchelf_path=None
    ) -> None:
        """Create a Patcher instance.

        :param str dynamic_linker: the path to the dynamic linker to set the
                                   elf file to.
        :param str root_path: the base path for the snap to determine
                              if use of $ORIGIN is possible.
        :param str preferred_patchelf_path: patch the necessary elf_files with
                                        this patchelf.
        """
        self._dynamic_linker = dynamic_linker
        self._root_path = root_path

        if preferred_patchelf_path:
            self._patchelf_cmd = preferred_patchelf_path
        else:
            self._patchelf_cmd = file_utils.get_snap_tool_path("patchelf")

        self._strip_cmd = file_utils.get_snap_tool_path("strip")

    def patch(self, *, elf_file: ElfFile) -> None:
        """Patch elf_file with the Patcher instance configuration.

        If the ELF is executable, patch it to use the configured linker.
        If the ELF has dependencies (DT_NEEDED), set an rpath to them.

        :param ElfFile elf: a data object representing an elf file and its
                            relevant attributes.
        :raises snapcraft.internal.errors.PatcherError:
            raised when the elf_file cannot be patched.
        """
        patchelf_args = []
        if elf_file.interp:
            patchelf_args.extend(["--set-interpreter", self._dynamic_linker])
        if elf_file.dependencies:
            rpath = self._get_rpath(elf_file)
            # Due to https://github.com/NixOS/patchelf/issues/94 we need
            # to first clear the current rpath
            self._run_patchelf(
                patchelf_args=["--remove-rpath"], elf_file_path=elf_file.path
            )
            # Parameters:
            # --force-rpath: use RPATH instead of RUNPATH.
            # --shrink-rpath: will remove unneeded entries, with the
            #                 side effect of preferring host libraries
            #                 so we simply do not use it.
            # --set-rpath: set the RPATH to the colon separated argument.
            patchelf_args.extend(["--force-rpath", "--set-rpath", rpath])

        # no patchelf_args means there is nothing to do.
        if not patchelf_args:
            return

        self._run_patchelf(patchelf_args=patchelf_args, elf_file_path=elf_file.path)

    def _run_patchelf(self, *, patchelf_args: List[str], elf_file_path: str) -> None:
        # Run patchelf on a copy of the primed file and replace it
        # after it is successful. This allows us to break the potential
        # hard link created when migrating the file across the steps of
        # the part.
        with tempfile.NamedTemporaryFile() as temp_file:
            shutil.copy2(elf_file_path, temp_file.name)

            cmd = [self._patchelf_cmd] + patchelf_args + [temp_file.name]
            try:
                logger.debug("executing: %s", " ".join(cmd))
                subprocess.check_call(cmd)
            # There is no need to catch FileNotFoundError as patchelf should be
            # bundled with snapcraft which means its lack of existence is a
            # "packager" error.
            except subprocess.CalledProcessError as call_error:
                raise errors.PatcherGenericError(
                    elf_file=elf_file_path, process_exception=call_error
                )

            # We unlink to break the potential hard link
            os.unlink(elf_file_path)
            shutil.copy2(temp_file.name, elf_file_path)

    def _get_existing_rpath(self, elf_file_path):
        output = subprocess.check_output(
            [self._patchelf_cmd, "--print-rpath", elf_file_path]
        )
        return output.decode().strip().split(":")

    def _get_rpath(self, elf_file) -> str:
        origin_rpaths = list()  # type: List[str]
        base_rpaths = set()  # type: Set[str]
        existing_rpaths = self._get_existing_rpath(elf_file.path)

        for dependency in elf_file.dependencies:
            if dependency.path:
                if dependency.in_base_snap:
                    base_rpaths.add(os.path.dirname(dependency.path))
                elif dependency.path.startswith(self._root_path):
                    rel_library_path = os.path.relpath(dependency.path, elf_file.path)
                    rel_library_path_dir = os.path.dirname(rel_library_path)
                    # return the dirname, with the first .. replace
                    # with $ORIGIN
                    origin_rpath = rel_library_path_dir.replace("..", "$ORIGIN", 1)
                    if origin_rpath not in origin_rpaths:
                        origin_rpaths.append(origin_rpath)

        if existing_rpaths:
            # Only keep those that mention origin and are not already in our
            # bundle.
            existing_rpaths = [
                r for r in existing_rpaths if "$ORIGIN" in r and r not in origin_rpaths
            ]
            origin_rpaths = existing_rpaths + origin_rpaths

        origin_paths = ":".join((r for r in origin_rpaths if r))
        core_base_rpaths = ":".join(sorted(base_rpaths))

        if origin_paths and core_base_rpaths:
            return "{}:{}".format(origin_paths, core_base_rpaths)
        elif origin_paths and not core_base_rpaths:
            return origin_paths
        else:
            return core_base_rpaths


def determine_ld_library_path(root: str) -> List[str]:
    """Determine additional library paths needed for the linker loader.

    This is a workaround until full library searching is implemented which
    works by searching for ld.so.conf in specific hard coded locations
    within root.

    :param root str: the root directory to search for specific ld.so.conf
                     entries.
    :returns: a list of strings of library paths where relevant libraries
              can be found within root.
    """
    # If more ld.so.conf files need to be supported, add them here.
    ld_config_globs = {"{}/usr/lib/*/mesa*/ld.so.conf".format(root)}

    ld_library_paths = []
    for this_glob in ld_config_globs:
        for ld_conf_file in glob.glob(this_glob):
            ld_library_paths.extend(_extract_ld_library_paths(ld_conf_file))

    return [root + path for path in ld_library_paths]


def _extract_ld_library_paths(ld_conf_file: str) -> List[str]:
    # From the ldconfig manpage, paths can be colon-, space-, tab-, newline-,
    # or comma-separated.
    path_delimiters = re.compile(r"[:\s,]")
    comments = re.compile(r"#.*$")

    paths = []
    with open(ld_conf_file, "r") as f:
        for line in f:
            # Remove comments from line
            line = comments.sub("", line).strip()

            if line:
                paths.extend(path_delimiters.split(line))

    return paths


_libraries = None


def get_elf_files(root: str, file_list: Sequence[str]) -> FrozenSet[ElfFile]:
    """Return a frozenset of elf files from file_list prepended with root.

    :param str root: the root directory from where the file_list is generated.
    :param file_list: a list of file in root.
    :returns: a frozentset of ElfFile objects.
    """
    elf_files = set()  # type: Set[ElfFile]

    for part_file in file_list:
        # Filter out object (*.o) files-- we only care about binaries.
        if part_file.endswith(".o"):
            continue

        # No need to crawl links-- the original should be here, too.
        path = os.path.join(root, part_file)  # type: str
        if os.path.islink(path):
            logger.debug("Skipped link {!r} while finding dependencies".format(path))
            continue

        # Ignore if file does not have ELF header.
        if not ElfFile.is_elf(path):
            continue

        try:
            elf_file = ElfFile(path=path)
        except elftools.common.exceptions.ELFError:
            # Ignore invalid ELF files.
            continue
        except errors.CorruptedElfFileError as exception:
            # Log if the ELF file seems corrupted
            logger.warning(exception.get_brief())
            continue

        # If ELF has dynamic symbols, add it.
        if elf_file.needed:
            elf_files.add(elf_file)

    return frozenset(elf_files)


def _get_dynamic_linker(library_list: List[str]) -> str:
    """Return the dynamic linker from library_list."""
    regex = re.compile(r"(?P<dynamic_linker>ld-[\d.]+.so)$")

    for library in library_list:
        m = regex.search(os.path.basename(library))
        if m:
            return library

    raise RuntimeError(
        "The format for the linker should be of the form "
        "<root>/ld-<X>.<Y>.so. There are no matches for the "
        "current libc6 package"
    )


def find_linker(*, root_path: str, snap_base_path: str) -> str:
    """Find and return the dynamic linker that would be seen at runtime.

    :param str root_path: the root path of a snap tree.
    :param str snap_base_path: absolute path to the snap once installed to
                               setup proper rpaths.
    :returns: the path to the dynamic linker to use
    """
    # We assume the current system will satisfy the GLIBC requirement,
    # get the current libc6 libraries (which includes the linker)
    libc6_libraries_list = repo.Repo.get_package_libraries("libc6")

    # For security reasons, we do not want to automatically pull in
    # libraries but expect them to be consciously brought in by stage-packages
    # instead.
    libc6_libraries_paths = [
        os.path.join(root_path, l[1:]) for l in libc6_libraries_list
    ]

    dynamic_linker = _get_dynamic_linker(libc6_libraries_paths)

    # Get the path to the "would be" dynamic linker when this snap is
    # installed. Strip the root_path from the retrieved dynamic_linker
    # variables + the leading `/` so that os.path.join can perform the
    # proper join with snap_base_path.
    dynamic_linker_path = os.path.join(
        snap_base_path, dynamic_linker[len(root_path) + 1 :]
    )

    return dynamic_linker_path



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

import fixtures
from testtools import TestCase
from testtools.matchers import Equals

from snapcraft.plugins.v2.npm import NpmPlugin


class NpmPluginTest(TestCase):
    def test_schema(self):
        schema = NpmPlugin.get_schema()

        self.assertThat(
            schema,
            Equals(
                {
                    "$schema": "http://json-schema.org/draft-04/schema#",
                    "additionalProperties": False,
                    "properties": {"npm-node-version": {"type": "string"}},
                    "required": ["source", "npm-node-version"],
                    "type": "object",
                }
            ),
        )

    def test_get_build_packages(self):
        plugin = NpmPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(plugin.get_build_packages(), Equals({"curl", "gcc"}))

    def test_get_build_environment(self):
        plugin = NpmPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(
            plugin.get_build_environment(),
            Equals({"PATH": "${SNAPCRAFT_PART_INSTALL}/bin:${PATH}"}),
        )

    def test_get_architecture_from_snap_arch(self):
        for snap_arch, node_arch in [
            ("amd64", "x64"),
            ("i386", "x86"),
            ("armhf", "armv7l"),
            ("arm64", "arm64"),
            ("ppc64el", "ppc64le"),
            ("s390x", "s390x"),
        ]:
            self.useFixture(fixtures.EnvironmentVariable("SNAP_ARCH", snap_arch))
            self.expectThat(NpmPlugin._get_architecture(), Equals(node_arch))

    def test_get_architecture_from_platform_for_x86(self):
        for platform_architecture, node_arch in [("32bit", "x86"), ("64bit", "x64")]:
            self.useFixture(
                fixtures.MockPatch("platform.machine", return_value="x86_64")
            )
            self.useFixture(
                fixtures.MockPatch(
                    "platform.architecture", return_value=("64bit", "ELF")
                )
            )
            self.expectThat(NpmPlugin._get_architecture(), Equals("x64"))

    def test_get_build_commands(self):
        self.useFixture(fixtures.EnvironmentVariable("SNAP_ARCH", "amd64"))

        class Options:
            npm_node_version = "6.0.0"

        plugin = NpmPlugin(part_name="my-part", options=Options())

        self.assertThat(
            plugin.get_build_commands(),
            Equals(
                [
                    dedent(
                        """\
                    if [ ! -f "${SNAPCRAFT_PART_INSTALL}/bin/node" ]; then
                        curl -s "https://nodejs.org/dist/v6.0.0/node-v6.0.0-linux-x64.tar.gz" | tar xzf - -C "${SNAPCRAFT_PART_INSTALL}/" --strip-components=1
                    fi
                    """
                    ),
                    'npm install -g --prefix "${SNAPCRAFT_PART_INSTALL}" $(npm pack . | tail -1)',
                ]
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.plugins.v1 import PluginV1


class LocalPlugin(PluginV1):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["foo"] = {"type": "string"}

        return schema

    @classmethod
    def get_pull_properties(cls):
        return ["foo", "stage-packages"]

    @classmethod
    def get_build_properties(cls):
        return ["foo", "stage-packages"]

    def build(self):
        return self.run(["touch", "build-stamp"], self.installdir)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Contains, Equals

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class ListTest(FakeStoreCommandsBaseTestCase):

    command_name = "list"

    def test_command_without_login_must_ask(self):
        # TODO: look into why this many calls are done inside snapcraft.storeapi
        self.fake_store_account_info.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            {"account_id": "abcd", "snaps": dict()},
            {"account_id": "abcd", "snaps": dict()},
            {"account_id": "abcd", "snaps": dict()},
        ]

        result = self.run_command(
            [self.command_name], input="user@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_list_empty(self):
        self.fake_store_account_info.mock.return_value = {
            "account_id": "abcd",
            "snaps": dict(),
        }

        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("There are no registered snaps."))

    def test_list_registered(self):
        self.command_name = "list-registered"
        self.fake_store_account_info.mock.return_value = {
            "account_id": "abcd",
            "snaps": dict(),
        }

        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("There are no registered snaps."))

    def test_registered(self):
        self.command_name = "registered"
        self.fake_store_account_info.mock.return_value = {
            "account_id": "abcd",
            "snaps": dict(),
        }

        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("There are no registered snaps."))

    def test_list_successfully(self):
        self.fake_store_account_info.mock.return_value = {
            "snaps": {
                "16": {
                    "foo": {
                        "status": "Approved",
                        "snap-id": "a_snap_id",
                        "private": False,
                        "since": "2016-12-12T01:01:01Z",
                        "price": "9.99",
                    },
                    "bar": {
                        "status": "ReviewPending",
                        "snap-id": "another_snap_id",
                        "private": True,
                        "since": "2016-12-12T01:01:01Z",
                        "price": None,
                    },
                    "baz": {
                        "status": "Approved",
                        "snap-id": "yet_another_snap_id",
                        "private": True,
                        "since": "2016-12-12T02:02:02Z",
                        "price": "6.66",
                    },
                    "boing": {
                        "status": "Approved",
                        "snap-id": "boing_snap_id",
                        "private": False,
                        "since": "2016-12-12T03:03:03Z",
                        "price": None,
                    },
                }
            }
        }

        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
            Name    Since                 Visibility    Price    Notes
            baz     2016-12-12T02:02:02Z  private       6.66     -
            boing   2016-12-12T03:03:03Z  public        -        -
            foo     2016-12-12T01:01:01Z  public        9.99     -"""
                )
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import textwrap

from testscenarios import multiply_scenarios

from snapcraft.internal.project_loader import find_extension, supported_extension_names

from .. import ProjectLoaderBaseTest


class ExtensionCombinationTest(ProjectLoaderBaseTest):
    scenarios = multiply_scenarios(
        *(
            [
                [
                    (
                        "{} extension".format(extension_name),
                        {"extension_{}".format(extension_name): extension_name},
                    ),
                    (
                        "no {} extension".format(extension_name),
                        {"extension_{}".format(extension_name): None},
                    ),
                ]
                for extension_name in supported_extension_names()
            ]
            or [[]]  # This guard can be removed once at least one extension is added
        )
    )

    def test_extensions_all_combinations_validate(self):
        common_bases = set(["core", "core18", "core20"])

        # Determine extension list given scenarios
        extension_names = list()
        for member_pair in inspect.getmembers(self):
            name = member_pair[0]
            if name.startswith("extension_"):
                value = getattr(self, name)
                if value:
                    extension_names.append(value)
                    ext = find_extension(value)
                    common_bases &= set(ext.get_supported_bases())

        # This shouldn't have any validation issues
        if common_bases:
            base = common_bases.pop()
            self.make_snapcraft_project(
                textwrap.dedent(
                    """\
                    name: test
                    version: "1"
                    summary: test
                    description: test
                    base: {base}
                    grade: stable
                    confinement: strict

                    apps:
                        test-app:
                            command: test-command
                            adapter: full
                            extensions: {extensions}

                    parts:
                        part1:
                            plugin: nil
                    """
                ).format(base=base, extensions=extension_names)
            )
        else:
            self.skipTest(f"no common base: {extension_names!r}")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import enum
import logging
import os
import tempfile
from typing import Any, Callable, Dict, List, Optional  # noqa: F401

from snapcraft import storeapi, yaml_utils
from snapcraft.internal import repo

logger = logging.getLogger(__name__)


class _SnapOp(enum.Enum):
    NOP = 0
    INJECT = 1
    INSTALL = 2
    REFRESH = 3


def _get_snap_channel(snap_name: str) -> storeapi.channels.Channel:
    """Returns the channel to use for snap_name."""
    env_channel = os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT", None)
    if env_channel is not None and snap_name == "snapcraft":
        channel = env_channel
        logger.warning(
            "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT is set: installing "
            "snapcraft from {}".format(channel)
        )
    else:
        channel = "latest/stable"

    return storeapi.channels.Channel(channel)


class _SnapManager:
    def __init__(
        self,
        *,
        snap_name: str,
        remote_snap_dir: str,
        latest_revision: Optional[str],
        inject_from_host: bool = True
    ) -> None:
        self.snap_name = snap_name
        self._remote_snap_dir = remote_snap_dir
        self._inject_from_host = inject_from_host

        self._latest_revision = latest_revision
        self.__required_operation = None  # type: Optional[_SnapOp]
        self.__repo = None  # type: Optional[repo.snaps.SnapPackage]
        self.__revision = None  # type: Optional[str]
        self.__install_cmd = None  # type: Optional[List[str]]
        self.__switch_cmd: Optional[List[str]] = None
        self.__assertion_ack_cmd = None  # type: Optional[List[str]]

    def _get_snap_repo(self):
        if self.__repo is None:
            self.__repo = repo.snaps.SnapPackage(self.snap_name)
        return self.__repo

    def get_op(self) -> _SnapOp:
        if self.__required_operation is not None:
            return self.__required_operation

        # From the point of view of multiple architectures if the target host (this)
        # is different than that of where these snaps run from, then we always need to
        # install from the store
        if self._inject_from_host:
            # Get information from the host.
            host_snap_repo = self._get_snap_repo()
            try:
                host_snap_info = host_snap_repo.get_local_snap_info()
                is_installed = host_snap_repo.installed
            except repo.errors.SnapdConnectionError:
                # This maybe because we are in a docker instance or another OS.
                is_installed = False
        else:
            is_installed = False

        # The evaluations for the required operation is as follows:
        # - if the snap is not installed on the host (is_installed == False),
        #   and the snap is not installed in the build environment
        #   (_latest_revision is None), then a store install will take place.
        # - else if the snap is not installed on the host (is_installed == False),
        #   but the is previously installed revision in the build environment
        #   (_latest_revision is not None), then a store install will take place.
        # - else if the snap is installed on the host (is_installed == True),
        #   and the snap installed in the build environment (_latest_revision) matches
        #   the one on the host, no operation takes place.
        # - else if the snap is installed on the host (is_installed == True),
        #   and the snap installed in the build environment (_latest_revision) does not
        #   match the one on the host, then a snap injection from the host will take place.
        if not is_installed and self._latest_revision is None:
            op = _SnapOp.INSTALL
        elif not is_installed and self._latest_revision is not None:
            op = _SnapOp.REFRESH
        elif is_installed and self._latest_revision == host_snap_info["revision"]:
            op = _SnapOp.NOP
        elif is_installed and self._latest_revision != host_snap_info["revision"]:
            op = _SnapOp.INJECT
        else:
            # This is a programmatic error
            raise RuntimeError(
                "Unhandled scenario for {!r} (host installed: {}, latest_revision {})".format(
                    self.snap_name, is_installed, self._latest_revision
                )
            )

        self.__required_operation = op
        return op

    def push_host_snap(self, *, file_pusher: Callable[..., None]) -> None:
        # TODO not being able to lock down on a snap revision can lead to races.
        host_snap_repo = self._get_snap_repo()
        with tempfile.TemporaryDirectory() as temp_dir:
            snap_file_path = os.path.join(temp_dir, "{}.snap".format(self.snap_name))
            assertion_file_path = os.path.join(
                temp_dir, "{}.assert".format(self.snap_name)
            )
            host_snap_repo.local_download(
                snap_path=snap_file_path, assertion_path=assertion_file_path
            )
            # Last item of __install_cmd holds the snap_file_path on the remote.
            file_pusher(
                source=snap_file_path, destination=self.get_snap_install_cmd()[-1]
            )
            # Last item of __assert_ack_cmd holds the snap_file_path on the remote.
            file_pusher(
                source=assertion_file_path, destination=self.get_assertion_ack_cmd()[-1]
            )

    def _set_data(self) -> None:
        op = self.get_op()
        host_snap_repo = self._get_snap_repo()

        install_cmd = list()  # type: List[str]
        switch_cmd: Optional[List[str]] = None
        assertion_ack_cmd = list()  # type: List[str]
        snap_revision = None

        if op == _SnapOp.INJECT:
            install_cmd = ["snap", "install"]
            host_snap_info = host_snap_repo.get_local_snap_info()
            snap_revision = host_snap_info["revision"]
            snap_channel = host_snap_info.get("tracking-channel")

            if not snap_revision.startswith("x") and snap_channel:
                switch_cmd = [
                    "snap",
                    "switch",
                    self.snap_name,
                    "--channel",
                    snap_channel,
                ]

            if snap_revision.startswith("x"):
                install_cmd.append("--dangerous")

            if host_snap_info["confinement"] == "classic":
                install_cmd.append("--classic")

            # File names need to be the last items in these two.
            install_cmd.append(
                os.path.join(
                    self._remote_snap_dir, "{}.snap".format(host_snap_repo.name)
                )
            )
            assertion_ack_cmd = [
                "snap",
                "ack",
                os.path.join(
                    self._remote_snap_dir, "{}.assert".format(host_snap_repo.name)
                ),
            ]

        elif op == _SnapOp.INSTALL or op == _SnapOp.REFRESH:
            install_cmd = ["snap", op.name.lower()]
            snap_channel = _get_snap_channel(self.snap_name)
            store_snap_info = storeapi.StoreClient().snap.get_info(self.snap_name)
            snap_channel_map = store_snap_info.get_channel_mapping(
                risk=snap_channel.risk, track=snap_channel.track
            )
            snap_revision = snap_channel_map.revision
            if snap_channel_map.confinement == "classic":
                install_cmd.append("--classic")
            install_cmd.extend(["--channel", snap_channel_map.channel_details.name])
            install_cmd.append(host_snap_repo.name)

        self.__install_cmd = install_cmd
        self.__switch_cmd = switch_cmd
        self.__assertion_ack_cmd = assertion_ack_cmd
        self.__revision = snap_revision

    def get_revision(self) -> str:
        if self.__revision is None:
            self._set_data()

        # Shouldn't happen - assert for sanity and mypy checking.
        if self.__revision is None:
            # Shouldn't happen.
            raise RuntimeError(
                "Unhandled scenario for {!r} (revision {})".format(
                    self.snap_name, self.__revision
                )
            )

        return self.__revision

    def get_snap_install_cmd(self) -> List[str]:
        if self.__install_cmd is None:
            self._set_data()

        # Shouldn't happen - assert for sanity and mypy checking.
        if self.__install_cmd is None:
            raise RuntimeError(
                "Unhandled scenario for {!r} (install_cmd {})".format(
                    self.snap_name, self.__install_cmd
                )
            )

        return self.__install_cmd

    def get_channel_switch_cmd(self) -> Optional[List[str]]:
        # The tracked channel can be None, so probe __revision instead
        if self.__revision is None:
            self._set_data()

        return self.__switch_cmd

    def get_assertion_ack_cmd(self) -> List[str]:
        if self.__assertion_ack_cmd is None:
            self._set_data()

        # Shouldn't happen - assert for sanity and mypy checking.
        if self.__assertion_ack_cmd is None:
            raise RuntimeError(
                "Unhandled scenario for {!r} (assertion_ack_cmd {})".format(
                    self.snap_name, self.__assertion_ack_cmd
                )
            )

        return self.__assertion_ack_cmd


def _load_registry(registry_filepath: Optional[str]) -> Dict[str, List[Any]]:
    if registry_filepath is None or not os.path.exists(registry_filepath):
        return dict()

    with open(registry_filepath) as registry_file:
        return yaml_utils.load(registry_file)


def _save_registry(
    registry_data: Dict[str, List[Any]], registry_filepath: Optional[str]
) -> None:
    if registry_filepath is None:
        return

    dirpath = os.path.dirname(registry_filepath)
    if dirpath:
        os.makedirs(dirpath, exist_ok=True)

    with open(registry_filepath, "w") as registry_file:
        yaml_utils.dump(registry_data, stream=registry_file)


class SnapInjector:
    """Handle the process of adding snaps into the build environment.

    The specific knowledge of the build environment where these snaps will
    be injected is not required, instead runnables to execute the required
    operations against the build environment are provided upon initialization.

    The snaps to install or refresh in the build environment are added by calling
    add and finally applied by calling apply. If an operation is required, the
    revision that eventually made it into the environment is recorded in the
    registry.
    """

    def __init__(
        self,
        *,
        registry_filepath: str,
        runner: Callable[..., Optional[bytes]],
        file_pusher: Callable[..., None],
        inject_from_host: bool = True
    ) -> None:
        """
        Initialize a SnapInjector instance.

        :param str registry_filepath: path to where recordings of previusly installed
                                      revisions of a snap can be queried and recorded.
        :param runner: a callable which can run commands in the build environment.
        :param file_pusher: a callable that can push file from the host into the build
                            environment.
        :param bool inject_from_host: whether to look for snaps on the host and inject them.
        """

        self._snaps = []  # type: List[_SnapManager]
        self._registry_filepath = registry_filepath
        self._inject_from_host = inject_from_host
        self._runner = runner
        self._file_pusher = file_pusher

        self._registry_data = _load_registry(registry_filepath)
        self._remote_snap_dir = "/var/tmp"

    def _disable_and_wait_for_refreshes(self) -> None:
        # Disable autorefresh for 1 day.
        hold_time = datetime.datetime.now() + datetime.timedelta(days=1)
        logger.debug("Holding refreshes for snaps.")
        self._runner(
            ["snap", "set", "system", "refresh.hold={}Z".format(hold_time.isoformat())],
            hide_output=True,
        )

        # Auto refresh may have kicked in while setting the hold.
        logger.debug("Waiting for pending snap auto refreshes.")
        self._runner(["snap", "watch", "--last=auto-refresh?"], hide_output=True)

    def _enable_snapd_snap(self) -> None:
        # Required to not install the core snap when building using
        # other bases.
        logger.debug("Enable use of snapd snap.")
        self._runner(
            ["snap", "set", "system", "experimental.snapd-snap=true"], hide_output=True
        )

    def _get_latest_revision(self, snap_name) -> Optional[str]:
        try:
            return self._registry_data[snap_name][-1]["revision"]
        except (IndexError, KeyError):
            return None

    def _record_revision(self, snap_name: str, snap_revision: str) -> None:
        entry = dict(revision=snap_revision)

        if snap_name not in self._registry_data:
            self._registry_data[snap_name] = [entry]
        else:
            self._registry_data[snap_name].append(entry)

    def add(self, snap_name: str) -> None:
        self._snaps.append(
            _SnapManager(
                snap_name=snap_name,
                remote_snap_dir=self._remote_snap_dir,
                latest_revision=self._get_latest_revision(snap_name),
                inject_from_host=self._inject_from_host,
            )
        )

    def apply(self) -> None:
        if all((s.get_op() == _SnapOp.NOP for s in self._snaps)):
            return

        # Allow using snapd from the snapd snap to leverage newer snapd features.
        if any(s.snap_name == "snapd" for s in self._snaps):
            self._enable_snapd_snap()

        # Disable refreshes so they do not interfere with installation ops.
        self._disable_and_wait_for_refreshes()

        # Filter out snaps with no operations.
        snaps = [snap for snap in self._snaps if snap.get_op() != _SnapOp.NOP]

        # Install snaps and assertions.
        for snap in snaps:
            if snap.get_op() == _SnapOp.INJECT:
                snap.push_host_snap(file_pusher=self._file_pusher)
                self._runner(snap.get_assertion_ack_cmd())
            self._runner(snap.get_snap_install_cmd())
            if snap.get_channel_switch_cmd() is not None:
                self._runner(snap.get_channel_switch_cmd())
            self._record_revision(snap.snap_name, snap.get_revision())

        _save_registry(self._registry_data, self._registry_filepath)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The nil plugin is useful for parts with no source.

Using this, parts can be defined purely by utilizing properties automatically
included by Snapcraft, e.g. stage-packages.
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class NilPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {},
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return set()

    def get_build_environment(self) -> Dict[str, str]:
        return dict()

    def get_build_commands(self) -> List[str]:
        return list()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from testtools.matchers import Equals

from snapcraft import formatting_utils
from tests import unit


class HumanizeListTestCases(unit.TestCase):
    def test_no_items(self):
        items = []
        output = formatting_utils.humanize_list(items, "and")
        self.assertThat(output, Equals(""))

    def test_one_item(self):
        items = ["foo"]
        output = formatting_utils.humanize_list(items, "and")
        self.assertThat(output, Equals("'foo'"))

    def test_two_items(self):
        items = ["foo", "bar"]
        output = formatting_utils.humanize_list(items, "and")
        self.assertThat(
            output,
            Equals("'bar' and 'foo'"),
            "Expected 'bar' before 'foo' due to sorting",
        )

    def test_three_items(self):
        items = ["foo", "bar", "baz"]
        output = formatting_utils.humanize_list(items, "and")
        self.assertThat(output, Equals("'bar', 'baz', and 'foo'"))

    def test_four_items(self):
        items = ["foo", "bar", "baz", "qux"]
        output = formatting_utils.humanize_list(items, "and")
        self.assertThat(output, Equals("'bar', 'baz', 'foo', and 'qux'"))

    def test_another_conjunction(self):
        items = ["foo", "bar", "baz", "qux"]
        output = formatting_utils.humanize_list(items, "or")
        self.assertThat(output, Equals("'bar', 'baz', 'foo', or 'qux'"))


def test_no_paths():
    with pytest.raises(ValueError):
        formatting_utils.format_path_variable("PATH", list(), "/usr", ":")


def test_one_path():
    paths = ["/bin"]
    output = formatting_utils.format_path_variable("PATH", paths, "/usr", ":")

    assert output == 'PATH="${PATH:+$PATH:}/usr/bin"'


def test_two_paths():
    paths = ["/bin", "/sbin"]
    output = formatting_utils.format_path_variable("PATH", paths, "/usr", ":")

    assert output == 'PATH="${PATH:+$PATH:}/usr/bin:/usr/sbin"'


def test_two_paths_other_paremeters():
    paths = ["/usr/bin", "/usr/sbin"]
    output = formatting_utils.format_path_variable("PATH", paths, "", ",")

    assert output == 'PATH="${PATH:+$PATH,}/usr/bin,/usr/sbin"'



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Equals, FileExists

from tests import integration


class YamlInSubdirTestCase(integration.TestCase):
    def test_stage(self):
        self.copy_project_to_cwd("yaml-in-subdir")
        os.chdir("subdir")
        self.run_snapcraft("stage")

        expected_file = os.path.join(self.stage_dir, "file")
        self.assertThat(expected_file, FileExists())
        with open(expected_file) as f:
            self.assertThat(f.read(), Equals("I'm a file\n"))

        expected_file = os.path.join(self.stage_dir, "subdirfile")
        self.assertThat(expected_file, FileExists())
        with open(expected_file) as f:
            self.assertThat(f.read(), Equals("I'm in the subdir\n"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from unittest.mock import Mock

import pytest
import tinydb

from snapcraft.internal.db import migration


@pytest.fixture
def test_db():
    db = tinydb.TinyDB(storage=tinydb.storages.MemoryStorage)
    return Mock(wraps=db)


@pytest.fixture(autouse=True)
def mock_datetime(monkeypatch):
    datetime_mock = Mock(wraps=datetime.datetime)
    datetime_mock.utcnow.return_value = datetime.datetime(2020, 1, 2, 3, 4, 5, 6)
    monkeypatch.setattr(migration, "datetime", datetime_mock)


def test_migration_on_empty_db(test_db):
    m = migration.MigrationV1(db=test_db, snapcraft_version="42")
    m.apply()

    assert test_db.table("control").all() == [
        {"created_with_snapcraft_version": "42", "schema_version": 1}
    ]
    assert test_db.table("migration").all() == [
        {
            "schema_version": 1,
            "snapcraft_version": "42",
            "timestamp": "2020-01-02T03:04:05.000006Z",
        }
    ]


def test_no_migration_on_v1_db(test_db):
    test_db.table("control").insert(
        {"created_with_snapcraft_version": "42", "schema_version": 1}
    )

    m = migration.MigrationV1(db=test_db, snapcraft_version="42")
    m.apply()

    assert test_db.table("control").all() == [
        {"created_with_snapcraft_version": "42", "schema_version": 1}
    ]
    assert test_db.table("migration").all() == []


def test_no_migration_on_v2_db(test_db):
    test_db.table("control").insert(
        {"created_with_snapcraft_version": "42", "schema_version": 2}
    )

    m = migration.MigrationV1(db=test_db, snapcraft_version="42")
    m.apply()

    assert test_db.table("control").all() == [
        {"created_with_snapcraft_version": "42", "schema_version": 2}
    ]
    assert test_db.table("migration").all() == []



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import pytest
from testtools.matchers import Equals

from snapcraft.internal import common, errors
from tests import unit


class CommonTestCase(unit.TestCase):
    def test_set_plugindir(self):
        plugindir = os.path.join(self.path, "testplugin")
        common.set_plugindir(plugindir)
        self.assertThat(plugindir, Equals(common.get_plugindir()))


def test_isurl():
    assert common.isurl("git://") is True
    assert common.isurl("bzr://") is True
    assert common.isurl("./") is False
    assert common.isurl("/foo") is False
    assert common.isurl("/foo:o") is False


class CommonMigratedTestCase(unit.TestCase):
    def test_parallel_build_count_migration_message(self):
        raised = self.assertRaises(
            errors.PluginOutdatedError, common.get_parallel_build_count
        )

        self.assertThat(
            str(raised), Equals("This plugin is outdated: use 'parallel_build_count'")
        )

    def test_deb_arch_migration_message(self):
        raised = self.assertRaises(errors.PluginOutdatedError, common.get_arch)

        self.assertThat(
            str(raised), Equals("This plugin is outdated: use 'project.deb_arch'")
        )

    def test_arch_triplet_migration_message(self):
        raised = self.assertRaises(errors.PluginOutdatedError, common.get_arch_triplet)

        self.assertThat(
            str(raised), Equals("This plugin is outdated: use 'project.arch_triplet'")
        )


class FormatInColumnsTestCase(unit.TestCase):

    elements_list = [
        "ant",
        "autotools",
        "catkin",
        "cmake",
        "copy",
        "go",
        "jdk",
        "kbuild",
        "kernel",
        "make",
        "maven",
        "nil",
        "nodejs",
        "python2",
        "python3",
        "scons",
        "tar-content",
    ]

    def test_format_output_in_columns_default(self):
        """Format output on 2 lines, with default max-width and space sep"""
        expected = [
            "ant        catkin  copy  jdk     kernel  maven  "
            "nodejs   python3  tar-content",
            "autotools  cmake   go    kbuild  make    nil    python2  scons  ",
        ]
        self.assertThat(
            common.format_output_in_columns(self.elements_list), Equals(expected)
        )

    def test_format_output_in_columns_narrow(self):
        """Format output on 3 lines, with narrow max-width and space sep"""
        expected = [
            "ant        cmake  jdk     make   nodejs   scons      ",
            "autotools  copy   kbuild  maven  python2  tar-content",
            "catkin     go     kernel  nil    python3",
        ]
        self.assertThat(
            common.format_output_in_columns(self.elements_list, max_width=60),
            Equals(expected),
        )

    def test_format_output_in_columns_large(self):
        """Format output on one big line, with default space sep"""
        expected = [
            "ant  autotools  catkin  cmake  copy  go  jdk  kbuild  "
            "kernel  make  maven  nil  nodejs  python2  python3  "
            "scons  tar-content"
        ]
        self.assertThat(
            common.format_output_in_columns(self.elements_list, max_width=990),
            Equals(expected),
        )

    def test_format_output_in_columns_one_space(self):
        """Format output with one space sep"""
        expected = [
            "ant       cmake jdk    make  nodejs  scons      ",
            "autotools copy  kbuild maven python2 tar-content",
            "catkin    go    kernel nil   python3",
        ]
        self.assertThat(
            common.format_output_in_columns(
                self.elements_list, max_width=60, num_col_spaces=1
            ),
            Equals(expected),
        )


class TestFormatSnapFileName:

    scenarios = [
        (
            "all info",
            dict(
                snap=dict(name="name", version="version", architectures=["amd64"]),
                expected="name_version_amd64.snap",
            ),
        ),
        (
            "missing version",
            dict(
                snap=dict(name="name", architectures=["amd64"]),
                expected="name_amd64.snap",
            ),
        ),
        (
            "no arch",
            dict(
                snap=dict(name="name", version="version"),
                expected="name_version_all.snap",
            ),
        ),
        (
            "multi",
            dict(
                snap=dict(
                    name="name", version="version", architectures=["amd64", "i386"]
                ),
                expected="name_version_multi.snap",
            ),
        ),
        (
            "pack",
            dict(
                snap=dict(name="name", version="version", arch=["amd64"]),
                expected="name_version_amd64.snap",
            ),
        ),
        (
            "pack multi",
            dict(
                snap=dict(name="name", version="version", arch=["amd64", "i386"]),
                expected="name_version_multi.snap",
            ),
        ),
    ]

    def test_filename(self, snap, expected):
        if "version" not in snap:
            snap_name = common.format_snap_name(snap, allow_empty_version=True)
        else:
            snap_name = common.format_snap_name(snap)

        assert snap_name == expected


def test_version_missing_and_not_allowed_is_error():
    # This is to not experience unexpected results given the
    # fact that version is not allowed.
    with pytest.raises(KeyError):
        common.format_snap_name(dict(name="name"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

import fixtures
from testtools.matchers import Equals, FileExists, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import crystal
from tests import unit

from . import PluginsV1BaseTestCase


class CrystalPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.fake_run = fixtures.MockPatch("snapcraft.internal.common.run")
        self.useFixture(self.fake_run)


class CrystalPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = crystal.CrystalPlugin.schema()

        properties = schema["properties"]
        for expected in ["crystal-channel"]:
            self.assertTrue(
                expected in properties,
                "Expected {!r} to be included in properties".format(expected),
            )

        # Check crystal-channel
        crystal_channel = properties["crystal-channel"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in crystal_channel,
                "Expected {!r} to be included in 'crystal-channel'".format(expected),
            )

        crystal_channel_type = crystal_channel["type"]
        self.assertThat(
            crystal_channel_type,
            Equals("string"),
            'Expected "crystal-channel" "type" to be "string", but '
            'it was "{}"'.format(crystal_channel_type),
        )

        crystal_channel_default = crystal_channel["default"]
        self.assertThat(
            crystal_channel_default,
            Equals("latest/stable"),
            'Expected "crystal-channel" "default" to be '
            '"latest/stable", but it was "{}"'.format(crystal_channel_default),
        )

        # Check crystal-build-options
        crystal_build_options = properties["crystal-build-options"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in crystal_channel,
                "Expected {!r} to be included in 'crystal-build-options'".format(
                    expected
                ),
            )

        crystal_build_options_type = crystal_channel["type"]
        self.assertThat(
            crystal_build_options_type,
            Equals("string"),
            'Expected "crystal-build-options" "type" to be "string", but '
            'it was "{}"'.format(crystal_build_options_type),
        )

        crystal_build_options_default = crystal_build_options["default"]
        self.assertThat(
            crystal_build_options_default,
            Equals([]),
            'Expected "crystal-build-options" "default" to be '
            '"[]", but it was "{}"'.format(crystal_build_options_default),
        )

        # Check required properties
        self.assertThat(schema["required"], Equals(["source"]))

    def test_get_pull_properties(self):
        expected_pull_properties = ["crystal-channel"]
        resulting_pull_properties = crystal.CrystalPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = ["crystal-build-options"]
        resulting_build_properties = crystal.CrystalPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


class MockElfFile:
    def __init__(self, *, path: str) -> None:
        self.path = path
        self.is_dynamic = True

    def load_dependencies(self, *args, **kwargs):
        return []


class CrystalPluginTest(CrystalPluginBaseTest):
    def test_pull(self):
        class Options:
            source = "dir"
            crystal_channel = "latest/stable"
            crystal_build_options = []

        plugin = crystal.CrystalPlugin("test-part", Options(), self.project)

        plugin.pull()

        self.fake_run.mock.assert_not_called()

    def test_build(self):
        class Options:
            source = "dir"
            crystal_channel = "latest/stable"
            crystal_build_options = []

        plugin = crystal.CrystalPlugin("test-part", Options(), self.project)

        # fake binaries being built
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.elf.ElfFile", side_effect=MockElfFile
            )
        )
        binaries = ["foo", "bar"]
        bin_dir = os.path.join(plugin.builddir, "bin")
        os.makedirs(bin_dir)
        for b in binaries:
            open(os.path.join(bin_dir, b), "w").close()

        plugin.build()

        self.fake_run.mock.assert_has_calls(
            [
                mock.call(["shards", "install", "--production"], cwd=plugin.builddir),
                mock.call(["shards", "build", "--production"], cwd=plugin.builddir),
            ]
        )
        self.assertThat(self.fake_run.mock.call_count, Equals(2))

        for b in binaries:
            self.assertThat(os.path.join(plugin.installdir, "bin", b), FileExists())

    def test_build_flags(self):
        class Options:
            source = "dir"
            crystal_channel = "latest/stable"
            crystal_build_options = ["-Dpreview_mt"]

        plugin = crystal.CrystalPlugin("test-part", Options(), self.project)

        # fake binaries being built
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.elf.ElfFile", side_effect=MockElfFile
            )
        )
        binaries = ["foo", "bar"]
        bin_dir = os.path.join(plugin.builddir, "bin")
        os.makedirs(bin_dir)
        for b in binaries:
            open(os.path.join(bin_dir, b), "w").close()

        plugin.build()

        self.fake_run.mock.assert_has_calls(
            [
                mock.call(["shards", "install", "--production"], cwd=plugin.builddir),
                mock.call(
                    ["shards", "build", "--production", "-Dpreview_mt"],
                    cwd=plugin.builddir,
                ),
            ]
        )
        self.assertThat(self.fake_run.mock.call_count, Equals(2))
        for b in binaries:
            self.assertThat(os.path.join(plugin.installdir, "bin", b), FileExists())

    def test_build_no_targets(self):
        class Options:
            source = "dir"
            crystal_channel = "latest/stable"
            crystal_build_options = []

        plugin = crystal.CrystalPlugin("test-part", Options(), self.project)

        self.assertRaises(errors.SnapcraftEnvironmentError, plugin.build)


class CrystalPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"
            crystal_channel = "latest/stable"
            crystal_build_options = []

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            crystal.CrystalPlugin,
            "test-part",
            self.options,
            self.project,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from textwrap import dedent

import yaml as yaml_utils
from testtools.matchers import Equals

from tests import fixture_setup, integration


class SetupPyMetadataTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()

        with open("setup.py", "w") as setup_file:
            print(
                dedent(
                    """\
                from setuptools import setup

                setup(
                    name='hello-world',
                    version='test-setuppy-version',
                    description='test-setuppy-description',
                    author='Canonical LTD',
                    author_email='snapcraft@lists.snapcraft.io',
                    scripts=['hello']
                )
                """
                ),
                file=setup_file,
            )

    def test_metadata_extracted_from_setuppy(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, version=None, description=None
        )
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part", {"plugin": "dump", "source": ".", "parse-info": ["setup.py"]}
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")
        snap_yaml_path = os.path.join("prime", "meta", "snap.yaml")
        with open(snap_yaml_path) as snap_yaml_file:
            snap_yaml = yaml_utils.load(snap_yaml_file)

        self.assertThat(snap_yaml["version"], Equals("test-setuppy-version"))
        self.assertThat(snap_yaml["description"], Equals("test-setuppy-description"))

    def test_all_metadata_from_yaml(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, version="test-yaml-version", description="test-yaml-description"
        )
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part", {"plugin": "dump", "source": ".", "parse-info": ["setup.py"]}
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")
        snap_yaml_path = os.path.join("prime", "meta", "snap.yaml")
        with open(snap_yaml_path) as snap_yaml_file:
            snap_yaml = yaml_utils.load(snap_yaml_file)

        self.assertThat(snap_yaml["version"], Equals("test-yaml-version"))
        self.assertThat(snap_yaml["description"], Equals("test-yaml-description"))

    def test_version_metadata_from_yaml_description_collected(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, version="test-yaml-version", description=None
        )
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part", {"plugin": "dump", "source": ".", "parse-info": ["setup.py"]}
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")
        snap_yaml_path = os.path.join("prime", "meta", "snap.yaml")
        with open(snap_yaml_path) as snap_yaml_file:
            snap_yaml = yaml_utils.load(snap_yaml_file)

        self.assertThat(snap_yaml["version"], Equals("test-yaml-version"))
        self.assertThat(snap_yaml["description"], Equals("test-setuppy-description"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019,2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The colcon plugin is useful for building ROS2 parts.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - colcon-packages:
      (list of strings)
      List of colcon packages to build. If not specified, all packages in the
      workspace will be built. If set to an empty list ([]), no packages will
      be built, which could be useful if you only want ROS debs in the snap.
    - colcon-packages-ignore:
      (list of strings)
      List of colcon packages to ignore. If not specified or set to an empty
      list ([]), no packages will be ignored.
    - colcon-source-space:
      (string)
      The source space containing colcon packages (defaults to 'src').
    - colcon-rosdistro:
      (string)
      The ROS distro to use. Available options are bouncy, crystal, dashing, and
      eloquent, all of which are only compatible with core18 as the base. The default
      value is crystal.
    - colcon-cmake-args:
      (list of strings)
      Arguments to pass to cmake projects. Note that any arguments here which match
      colcon arguments need to be prefixed with a space. This can be done by quoting
      each argument with a leading space.
    - colcon-catkin-cmake-args:
      (list of strings)
      Arguments to pass to catkin packages. Note that any arguments here which match
      colcon arguments need to be prefixed with a space. This can be done by quoting
      each argument with a leading space.
    - colcon-ament-cmake-args:
      (list of strings)
      Arguments to pass to ament_cmake packages. Note that any arguments here which
      match colcon arguments need to be prefixed with a space. This can be done by
      quoting each argument with a leading space.
"""

import collections
import contextlib
import logging
import os
import pathlib
import re
import shutil
import textwrap
from typing import List

from snapcraft import file_utils
from snapcraft.internal import errors, mangling, os_release, repo
from snapcraft.internal.meta.package_repository import (
    PackageRepository,
    PackageRepositoryApt,
)
from snapcraft.plugins.v1 import PluginV1, _python, _ros

logger = logging.getLogger(__name__)

# Map bases to ROS releases
_ROSDISTRO_TO_BASE_MAP = {
    "bouncy": "core18",
    "crystal": "core18",
    "dashing": "core18",
    "eloquent": "core18",
}

# Snaps can still be built with ROS distros that are end-of-life, but such
# things are not supported. Maintain a list so we can warn about such things.
# This really should be using rosdistro to automatically detect the support
# status, but that's a larger feature than we want to implement at this time.
_EOL_ROSDISTROS = ["bouncy", "crystal"]

# Map bases to Ubuntu releases. Every base in _ROSDISTRO_TO_BASE_MAP needs to be
# specified here.
_BASE_TO_UBUNTU_RELEASE_MAP = {"core18": "bionic"}

_SUPPORTED_DEPENDENCY_TYPES = {"apt", "pip"}


class ColconInvalidSystemDependencyError(errors.SnapcraftError):
    fmt = (
        "Package {dependency!r} isn't a valid system dependency. Did you "
        "forget to add it to colcon-packages? If not, add the Ubuntu package "
        "containing it to stage-packages until you can get it into the rosdep "
        "database."
    )

    def __init__(self, dependency):
        super().__init__(dependency=dependency)


class ColconUnsupportedDependencyTypeError(errors.SnapcraftError):
    fmt = (
        "Package {dependency!r} resolved to an unsupported type of "
        "dependency: {dependency_type!r}."
    )

    def __init__(self, dependency_type, dependency):
        super().__init__(dependency_type=dependency_type, dependency=dependency)


class ColconWorkspaceIsRootError(errors.SnapcraftError):
    fmt = (
        "colcon-source-space cannot be the root of the colcon workspace; use a "
        "subdirectory."
    )


class ColconAptDependencyFetchError(errors.SnapcraftError):
    fmt = "Failed to fetch apt dependencies: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class ColconPackagePathNotFoundError(errors.SnapcraftError):
    fmt = "Failed to find package path: {path!r}"

    def __init__(self, path):
        super().__init__(path=path)


class ColconPluginBaseError(errors.PluginBaseError):
    fmt = (
        "The colcon plugin (used by part {part_name!r}) does not support using base "
        "{base!r} with rosdistro {rosdistro!r}."
    )

    def __init__(self, part_name, base, rosdistro):
        super(errors.PluginBaseError, self).__init__(
            part_name=part_name, base=base, rosdistro=rosdistro
        )


class ColconPlugin(PluginV1):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["colcon-rosdistro"] = {
            "type": "string",
            "default": "crystal",
            "enum": list(_ROSDISTRO_TO_BASE_MAP.keys()),
        }

        schema["properties"]["colcon-packages"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
        }

        schema["properties"]["colcon-source-space"] = {
            "type": "string",
            "default": "src",
        }

        schema["properties"]["colcon-cmake-args"] = {
            "type": "array",
            "minitems": 1,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["colcon-catkin-cmake-args"] = {
            "type": "array",
            "minitems": 1,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["colcon-ament-cmake-args"] = {
            "type": "array",
            "minitems": 1,
            "items": {"type": "string"},
            "default": [],
        }

        schema["properties"]["colcon-packages-ignore"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["colcon-packages", "colcon-source-space", "colcon-rosdistro"]

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return [
            "colcon-cmake-args",
            "colcon-catkin-cmake-args",
            "colcon-ament-cmake-args",
            "colcon-packages-ignore",
        ]

    @classmethod
    def get_required_package_repositories(cls) -> List[PackageRepository]:
        codename = os_release.OsRelease().version_codename()
        return [
            PackageRepositoryApt(
                formats=["deb"],
                components=["main"],
                key_id="C1CF6E31E6BADE8868B172B4F42ED6FBAB17C654",
                url="http://packages.ros.org/ros2/ubuntu",
                suites=[codename],
            )
        ]

    @property
    def _pip(self):
        if not self.__pip:
            self.__pip = _python.Pip(
                python_major_version="3",  # ROS2 uses python3
                part_dir=self.partdir,
                install_dir=self.installdir,
                stage_dir=self.project.stage_dir,
            )

        return self.__pip

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self.out_of_source_build = True

        self._rosdistro = options.colcon_rosdistro
        if project._get_build_base() != _ROSDISTRO_TO_BASE_MAP[self._rosdistro]:
            raise ColconPluginBaseError(
                self.name, project._get_build_base(), self._rosdistro
            )

        if self._rosdistro in _EOL_ROSDISTROS:
            logger.warning(
                "The {!r} ROS distro has reached end-of-life and is no longer supported. Use at your own risk.".format(
                    self._rosdistro
                )
            )

        # Beta warning. Remove this comment and warning once plugin is stable.
        logger.warning(
            "The colcon plugin is currently in beta, its API may break. Use at your "
            "own risk."
        )

        self.__pip = None
        self._rosdep_path = os.path.join(self.partdir, "rosdep")
        self._ros_underlay = os.path.join(
            self.installdir, "opt", "ros", self._rosdistro
        )

        # Colcon packages cannot be installed into the same workspace as upstream ROS
        # components, which are Ament. We need to install them alongside and chain the
        # workspaces (upstream is the underlay, stuff we're building here is the
        # overlay).
        self._ros_overlay = os.path.join(self.installdir, "opt", "ros", "snap")

        # Always fetch colcon in order to build the workspace
        self.stage_packages.append("python3-colcon-common-extensions")

        # Get a unique set of packages
        self._packages = None
        if options.colcon_packages is not None:
            self._packages = set(options.colcon_packages)

        # The path created via the `source` key (or a combination of `source`
        # and `source-subdir` keys) needs to point to a valid Colcon workspace
        # containing another subdirectory called the "source space." By
        # default, this is a directory named "src," but it can be remapped via
        # the `source-space` key. It's important that the source space is not
        # the root of the Colcon workspace, since Colcon won't work that way
        # and it'll create a circular link that causes rosdep to hang.
        if self.options.source_subdir:
            self._ros_package_path = os.path.join(
                self.sourcedir,
                self.options.source_subdir,
                self.options.colcon_source_space,
            )
        else:
            self._ros_package_path = os.path.join(
                self.sourcedir, self.options.colcon_source_space
            )

        if os.path.abspath(self.sourcedir) == os.path.abspath(self._ros_package_path):
            raise ColconWorkspaceIsRootError()

        self.stage_packages_path = pathlib.Path(self.partdir) / "colcon_stage_packages"

    def env(self, root):
        """Runtime environment for ROS binaries and services."""

        env = [
            'AMENT_PYTHON_EXECUTABLE="{}"'.format(
                os.path.join(root, "usr", "bin", "python3")
            ),
            'COLCON_PYTHON_EXECUTABLE="{}"'.format(
                os.path.join(root, "usr", "bin", "python3")
            ),
            'SNAP_COLCON_ROOT="{}"'.format(root),
        ]

        # Each of these lines is prepended with an `export` when the environment is
        # actually generated. In order to inject real shell code we have to hack it in
        # by appending it on the end of an item already in the environment.
        # FIXME: There should be a better way to do this. LP: #1792034
        env[-1] = env[-1] + "\n\n" + self._source_setup_sh(root)

        return env

    def pull(self):
        """Copy source into build directory and fetch dependencies.

        Colcon packages can specify their system dependencies in their
        package.xml. In order to support that, the Colcon packages are
        interrogated for their dependencies here. Since `stage-packages` are
        already installed by the time this function is run, the dependencies
        from the package.xml are pulled down explicitly.
        """

        super().pull()

        # Make sure the package path exists before continuing. We only care about doing
        # this if there are actually packages to build, which is indicated both by
        # self._packages being None as well as a non-empty list.
        packages_to_build = self._packages is None or len(self._packages) > 0
        if packages_to_build and not os.path.exists(self._ros_package_path):
            raise ColconPackagePathNotFoundError(self._ros_package_path)

        # Use rosdep for dependency detection and resolution
        rosdep = _ros.rosdep.Rosdep(
            ros_distro=self._rosdistro,
            ros_version="2",
            ros_package_path=self._ros_package_path,
            rosdep_path=self._rosdep_path,
            ubuntu_distro=_BASE_TO_UBUNTU_RELEASE_MAP[self.project._get_build_base()],
            base=self.project._get_build_base(),
            target_arch=self.project._get_stage_packages_target_arch(),
        )
        rosdep.setup()

        self._setup_dependencies(rosdep)

    def _setup_dependencies(self, rosdep):
        # Parse the Colcon packages to pull out their system dependencies
        system_dependencies = _find_system_dependencies(self._packages, rosdep)

        # Pull down and install any apt dependencies that were discovered
        self._setup_apt_dependencies(system_dependencies.get("apt"))

        # Pull down and install any pip dependencies that were discovered
        self._setup_pip_dependencies(system_dependencies.get("pip"))

    def _setup_apt_dependencies(self, apt_dependencies):
        if not apt_dependencies:
            return

        logger.info("Installing apt dependencies...")
        try:
            repo.Ubuntu.fetch_stage_packages(
                package_names=apt_dependencies,
                stage_packages_path=self.stage_packages_path,
                base=self.project._get_build_base(),
                target_arch=self.project._get_stage_packages_target_arch(),
            )
        except repo.errors.PackageNotFoundError as e:
            raise ColconAptDependencyFetchError(e.message)
        repo.Ubuntu.unpack_stage_packages(
            stage_packages_path=self.stage_packages_path,
            install_path=pathlib.Path(self.installdir),
        )

    def _setup_pip_dependencies(self, pip_dependencies):
        if pip_dependencies:
            self._pip.setup()

            logger.info("Fetching pip dependencies...")
            self._pip.download(pip_dependencies)

            logger.info("Installing pip dependencies...")
            self._pip.install(pip_dependencies)

    def clean_pull(self):
        super().clean_pull()

        # Remove the rosdep path, if any
        with contextlib.suppress(FileNotFoundError):
            shutil.rmtree(self._rosdep_path)

        with contextlib.suppress(FileNotFoundError):
            shutil.rmtree(self.stage_packages_path)

        # Clean pip packages, if any
        self._pip.clean_packages()

    def _source_setup_sh(self, root):
        underlaydir = os.path.join(root, "opt", "ros", self._rosdistro)
        overlaydir = os.path.join(root, "opt", "ros", "snap")

        source_script = textwrap.dedent(
            """
            # First, source the upstream ROS underlay
            if [ -f "{underlay_setup}" ]; then
                . "{underlay_setup}"
            fi

            # Then source the overlay
            if [ -f "{overlay_setup}" ]; then
                . "{overlay_setup}"
            fi
        """
        ).format(
            underlay_setup=os.path.join(underlaydir, "setup.sh"),
            overlay_setup=os.path.join(overlaydir, "local_setup.sh"),
        )

        # We need to source ROS's setup.sh at this point. However, it accepts
        # arguments (thus will parse $@), and we really don't want it to, since
        # $@ in this context will be meant for the app being launched
        # (LP: #1660852). So we'll backup all args, source the setup.sh, then
        # restore all args for the wrapper's `exec` line.
        return textwrap.dedent(
            """
            # Shell quote arbitrary string by replacing every occurrence of '
            # with '\\'', then put ' at the beginning and end of the string.
            # Prepare yourself, fun regex ahead.
            quote()
            {{
                for i; do
                    printf %s\\\\n "$i" | sed "s/\'/\'\\\\\\\\\'\'/g;1s/^/\'/;\$s/\$/\' \\\\\\\\/"
                done
                echo " "
            }}

            BACKUP_ARGS=$(quote "$@")
            set --
            {}
            eval "set -- $BACKUP_ARGS"
        """  # noqa: W605
        ).format(
            source_script
        )  # noqa

    def build(self):
        """Build Colcon packages.

        This function runs some pre-build steps to prepare the sources for building in
        the Snapcraft environment, builds the packages with colcon, and finally runs
        some post-build clean steps to prepare the newly-minted install to be packaged
        as a .snap.
        """

        super().build()

        logger.info("Preparing to build colcon packages...")
        self._prepare_build()

        logger.info("Building colcon packages...")
        self._build_colcon_packages()

        logger.info("Cleaning up newly installed colcon packages...")
        self._finish_build()

    def _prepare_build(self):
        # Fix all shebangs to use the in-snap python.
        mangling.rewrite_python_shebangs(self.installdir)

        # Rewrite the prefixes to point to the in-part rosdir instead of the system
        self._fix_prefixes()

        # Each Colcon package distributes .cmake files so they can be found via
        # find_package(). However, the Ubuntu packages pulled down as
        # dependencies contain .cmake files pointing to system paths (e.g.
        # /usr/lib, /usr/include, etc.). They need to be rewritten to point to
        # the install directory.
        def _new_path(path):
            if not path.startswith(self.installdir):
                # Not using os.path.join here as `path` is absolute.
                return self.installdir + path
            return path

        self._rewrite_cmake_paths(_new_path)

    def _rewrite_cmake_paths(self, new_path_callable):
        def _rewrite_paths(match):
            paths = match.group(1).strip().split(";")
            for i, path in enumerate(paths):
                # Offer the opportunity to rewrite this path if it's absolute.
                if os.path.isabs(path):
                    paths[i] = new_path_callable(path)

            return '"' + ";".join(paths) + '"'

        # Looking for any path-like string
        file_utils.replace_in_file(
            self._ros_underlay,
            re.compile(r".*Config.cmake$"),
            re.compile(r'"(.*?/.*?)"'),
            _rewrite_paths,
        )

        file_utils.replace_in_file(
            os.path.join(self.installdir, "usr", "lib", "cmake"),
            re.compile(r".*.cmake$"),
            re.compile(r'"(.*?/.*?)"'),
            _rewrite_paths,
        )

    def _finish_build(self):
        # Fix all shebangs to use the in-snap python.
        mangling.rewrite_python_shebangs(self.installdir)

        # We've finished the build, but we need to make sure we turn the cmake
        # files back into something that doesn't include our installdir. This
        # way it's usable from the staging area, and won't clash with the same
        # file coming from other parts.
        pattern = re.compile(r"^{}".format(self.installdir))

        def _new_path(path):
            return pattern.sub("$ENV{SNAPCRAFT_STAGE}", path)

        self._rewrite_cmake_paths(_new_path)

        # Rewrite prefixes for both the underlay and overlay.
        self._fix_prefixes()

        # If pip dependencies were installed, generate a sitecustomize that
        # allows access to them.
        if self._pip.is_setup() and self._pip.list(user=True):
            _python.generate_sitecustomize(
                "3", stage_dir=self.project.stage_dir, install_dir=self.installdir
            )

    def _fix_prefixes(self):
        installdir_pattern = re.compile(r"^{}".format(self.installdir))
        new_prefix = "$SNAP_COLCON_ROOT"

        def _rewrite_prefix(match):
            # Group 1 is the variable definition, group 2 is the path, which we may need
            # to modify.
            path = match.group(3).strip(" \n\t'\"")

            # Bail early if this isn't even a path, or if it's already been rewritten
            if os.path.sep not in path or new_prefix in path:
                return match.group()

            # If the path doesn't start with the installdir, then it needs to point to
            # the underlay given that the upstream ROS packages are expecting to be in
            # /opt/ros/.
            if not path.startswith(self.installdir):
                path = os.path.join(new_prefix, path.lstrip("/"))

            return match.expand(
                '\\1\\2"{}"\\4'.format(installdir_pattern.sub(new_prefix, path))
            )

        # Set the AMENT_CURRENT_PREFIX throughout to the in-snap prefix
        file_utils.replace_in_file(
            self.installdir,
            re.compile(r""),
            re.compile(r"(\${)(AMENT_CURRENT_PREFIX:=)(.*)(})"),
            _rewrite_prefix,
        )

        # Set the COLCON_CURRENT_PREFIX (if it's in the installdir) to the in-snap
        # prefix
        file_utils.replace_in_file(
            self.installdir,
            re.compile(r""),
            re.compile(
                r"()(COLCON_CURRENT_PREFIX=)(['\"].*{}.*)()".format(self.installdir)
            ),
            _rewrite_prefix,
        )

        # Set the _colcon_prefix_sh_COLCON_CURRENT_PREFIX throughout to the in-snap
        # prefix
        file_utils.replace_in_file(
            self.installdir,
            re.compile(r""),
            re.compile(r"()(_colcon_prefix_sh_COLCON_CURRENT_PREFIX=)(.*)()"),
            _rewrite_prefix,
        )

        # Set the _colcon_package_sh_COLCON_CURRENT_PREFIX throughout to the in-snap
        # prefix
        file_utils.replace_in_file(
            self.installdir,
            re.compile(r""),
            re.compile(r"()(_colcon_package_sh_COLCON_CURRENT_PREFIX=)(.*)()"),
            _rewrite_prefix,
        )

        # Set the _colcon_prefix_chain_sh_COLCON_CURRENT_PREFIX throughout to the in-snap
        # prefix
        file_utils.replace_in_file(
            self.installdir,
            re.compile(r""),
            re.compile(r"()(_colcon_prefix_chain_sh_COLCON_CURRENT_PREFIX=)(.*)()"),
            _rewrite_prefix,
        )

        # Set the _colcon_python_executable throughout to use the in-snap python
        file_utils.replace_in_file(
            self.installdir,
            re.compile(r""),
            re.compile(r"()(_colcon_python_executable=)(.*)()"),
            _rewrite_prefix,
        )

    def _build_colcon_packages(self):
        # Nothing to do if no packages were specified
        if self._packages is not None and len(self._packages) == 0:
            return

        colconcmd = ["colcon", "build", "--merge-install"]

        if self._packages:
            # Specify the packages to be built
            colconcmd.append("--packages-select")
            colconcmd.extend(self._packages)

        if self.options.colcon_packages_ignore:
            colconcmd.extend(
                ["--packages-ignore"] + self.options.colcon_packages_ignore
            )

        # Don't clutter the real ROS workspace-- use the Snapcraft build
        # directory
        colconcmd.extend(["--build-base", self.builddir])

        # Account for a non-default source space by always specifying it
        colconcmd.extend(["--base-paths", self._ros_package_path])

        # Specify that the packages should be installed into the overlay
        colconcmd.extend(["--install-base", self._ros_overlay])

        # Specify the number of workers
        colconcmd.append("--parallel-workers={}".format(self.parallel_build_count))

        # All the arguments that follow are meant for CMake
        colconcmd.append("--cmake-args")

        build_type = "Release"
        if "debug" in self.options.build_attributes:
            build_type = "Debug"
        colconcmd.extend(["-DCMAKE_BUILD_TYPE={}".format(build_type)])

        # Finally, add any cmake-args requested from the plugin options
        colconcmd.extend(self.options.colcon_cmake_args)

        if self.options.colcon_catkin_cmake_args:
            colconcmd.extend(
                ["--catkin-cmake-args"] + self.options.colcon_catkin_cmake_args
            )

        if self.options.colcon_ament_cmake_args:
            colconcmd.extend(
                ["--ament-cmake-args"] + self.options.colcon_ament_cmake_args
            )

        self.run(colconcmd)


def _find_system_dependencies(colcon_packages, rosdep):
    """Find system dependencies for a given set of Colcon packages."""

    resolved_dependencies = {}
    dependencies = set()

    logger.info("Determining system dependencies for Colcon packages...")
    if colcon_packages is not None:
        for package in colcon_packages:
            # Query rosdep for the list of dependencies for this package
            dependencies |= rosdep.get_dependencies(package)
    else:
        # Rather than getting dependencies for an explicit list of packages,
        # let's get the dependencies for the entire workspace.
        dependencies |= rosdep.get_dependencies()

    for dependency in dependencies:
        _resolve_package_dependencies(
            colcon_packages, dependency, rosdep, resolved_dependencies
        )

    # We currently have nested dict structure of:
    #    dependency name -> package type -> package names
    #
    # We want to return a flattened dict of package type -> package names.
    flattened_dependencies = collections.defaultdict(set)
    for dependency_types in resolved_dependencies.values():
        for key, value in dependency_types.items():
            flattened_dependencies[key] |= value

    # Finally, return that dict of dependencies
    return flattened_dependencies


def _resolve_package_dependencies(
    colcon_packages, dependency, rosdep, resolved_dependencies
):
    # No need to resolve this dependency if we know it's local, or if
    # we've already resolved it into a system dependency
    if dependency in resolved_dependencies or (
        colcon_packages and dependency in colcon_packages
    ):
        return

    # In this situation, the package depends on something that we
    # weren't instructed to build. It's probably a system dependency,
    # but the developer could have also forgotten to tell us to build
    # it.
    try:
        these_dependencies = rosdep.resolve_dependency(dependency)
    except _ros.rosdep.RosdepDependencyNotResolvedError:
        raise ColconInvalidSystemDependencyError(dependency)

    for key, value in these_dependencies.items():
        if key not in _SUPPORTED_DEPENDENCY_TYPES:
            raise ColconUnsupportedDependencyTypeError(key, dependency)

        resolved_dependencies[dependency] = {key: value}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import scons
from tests import unit

from . import PluginsV1BaseTestCase


class SconsPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        """Test validity of the Scons Plugin schema"""
        schema = scons.SconsPlugin.schema()

        # Verify the presence of all properties
        properties = schema["properties"]
        self.assertTrue(
            "scons-options" in properties,
            'Expected "scons-options" to be included in ' "properties",
        )

        scons_options = properties["scons-options"]

        self.assertTrue(
            "type" in scons_options, 'Expected "type" to be included in "scons-options"'
        )
        self.assertThat(
            scons_options["type"],
            Equals("array"),
            'Expected "scons-options" "type" to be "array", but '
            'it was "{}"'.format(scons_options["type"]),
        )

        self.assertTrue(
            "minitems" in scons_options,
            'Expected "minitems" to be included in "scons-options"',
        )
        self.assertThat(
            scons_options["minitems"],
            Equals(1),
            'Expected "scons-options" "minitems" to be 1, but '
            'it was "{}"'.format(scons_options["minitems"]),
        )

        self.assertTrue(
            "uniqueItems" in scons_options,
            'Expected "uniqueItems" to be included in "scons-options"',
        )
        self.assertTrue(
            scons_options["uniqueItems"],
            'Expected "scons-options" "uniqueItems" to be "True"',
        )

    def test_get_pull_properties(self):
        expected_pull_properties = []
        resulting_pull_properties = scons.SconsPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = ["scons-options"]
        resulting_build_properties = scons.SconsPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


class SconsPluginTest(PluginsV1BaseTestCase):
    """Plugin to provide snapcraft support for the scons build system"""

    def setUp(self):
        super().setUp()

        class Options:
            """Internal Options Class matching the Scons plugin"""

            scons_options = ["--debug=explain"]

        self.options = Options()

    def scons_build(self):
        """Helper to call a full build"""
        plugin = scons.SconsPlugin("test-part", self.options, self.project)
        # Create fake scons
        plugin.build()
        return plugin

    @mock.patch.object(scons.SconsPlugin, "run")
    def test_build_with_destdir(self, run_mock):
        """Test building via scons and check for known calls and destdir"""
        plugin = self.scons_build()
        env = os.environ.copy()
        env["DESTDIR"] = plugin.installdir

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["scons", "--debug=explain"]),
                mock.call(["scons", "install", "--debug=explain"], env=env),
            ]
        )


class SconsPluginUnsupportedBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            scons.SconsPlugin,
            "test-part",
            self.options,
            self.project,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum
import logging
from collections import OrderedDict
from typing import Any, Dict

from snapcraft.internal.meta import errors

logger = logging.getLogger(__name__)


@enum.unique
class SystemUserScope(enum.Enum):
    SHARED = 1


class SystemUser:
    def __init__(self, *, name: str, scope: SystemUserScope) -> None:
        self.name = name
        self.scope = scope

    @classmethod
    def from_dict(cls, *, user_name: str, user_dict: Dict[str, str]) -> "SystemUser":
        raw_scope = user_dict.get("scope", None)
        if raw_scope is None:
            raise errors.SystemUsernamesValidationError(
                name=user_name, message="'scope' is undefined"
            )

        try:
            scope = SystemUserScope[raw_scope.upper()]
        except KeyError:
            raise errors.SystemUsernamesValidationError(
                name=user_name, message=f"scope {raw_scope!r} is invalid"
            )

        return SystemUser(name=user_name, scope=scope)

    @classmethod
    def from_object(cls, *, user_object: Any, user_name: str) -> "SystemUser":
        if user_object is None:
            raise errors.SystemUsernamesValidationError(
                name=user_name, message="undefined user"
            )
        elif isinstance(user_object, str):
            try:
                scope = SystemUserScope[user_object.upper()]
            except KeyError:
                raise errors.SystemUsernamesValidationError(
                    name=user_name, message=f"scope {user_object!r} is invalid"
                )

            return SystemUser(name=user_name, scope=scope)
        elif isinstance(user_object, dict):
            return cls.from_dict(user_dict=user_object, user_name=user_name)

        raise errors.SystemUsernamesValidationError(
            name=user_name, message=f"unknown syntax for {user_object!r}"
        )

    def to_dict(self):
        user_dict = OrderedDict()
        user_dict["scope"] = self.scope.name.lower()
        return user_dict

    def validate(self):
        if not isinstance(self.scope, SystemUserScope):
            raise errors.SystemUsernamesValidationError(
                name=self.name, message=f"scope {self.scope!r} is invalid"
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from collections import OrderedDict
from typing import Any, Dict, List, Optional

from snapcraft.internal.meta.errors import HookValidationError


class Hook:
    """Representation of a generic snap hook."""

    def __init__(
        self,
        *,
        hook_name: str,
        command_chain: Optional[List[str]] = None,
        plugs: Optional[List[str]] = None,
        passthrough: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._hook_name = hook_name

        self.command_chain: List[str] = list()
        if command_chain:
            self.command_chain = command_chain

        self.plugs: List[str] = list()
        if plugs:
            self.plugs = plugs

        self.passthrough: Dict[str, Any] = dict()
        if passthrough:
            self.passthrough = passthrough

    @property
    def hook_name(self) -> str:
        """Read-only to ensure consistency with Snap dictionary mappings."""

        return self._hook_name

    def _validate_name(self) -> None:
        """Validate hook name."""

        if not re.match("^[a-z](?:-?[a-z0-9])*$", self.hook_name):
            raise HookValidationError(
                hook_name=self.hook_name,
                message="{!r} is not a valid hook name. Hook names consist of lower-case alphanumeric characters and hyphens. They cannot start or end with a hyphen.".format(
                    self.hook_name
                ),
            )

    def _validate_command_chain(self) -> None:
        """Validate command-chain names."""

        # Would normally get caught/handled by schema validation.
        for command in self.command_chain:
            if not re.match("^[A-Za-z0-9/._#:$-]*$", command):
                raise HookValidationError(
                    hook_name=self.hook_name,
                    message=f"{command!r} is not a valid command-chain command.",
                )

    def validate(self) -> None:
        """Validate hook, raising exception if invalid."""

        self._validate_name()
        self._validate_command_chain()

    @classmethod
    def from_dict(cls, hook_dict: Dict[str, Any], hook_name: str) -> "Hook":
        """Create hook from dictionary."""

        return Hook(
            hook_name=hook_name,
            command_chain=hook_dict.get("command-chain", None),
            plugs=hook_dict.get("plugs", None),
            passthrough=hook_dict.get("passthrough", None),
        )

    def to_dict(self) -> Dict[str, Any]:
        """Create dictionary from hook."""

        hook_dict: Dict[str, Any] = OrderedDict()

        if self.command_chain:
            hook_dict["command-chain"] = self.command_chain

        if self.plugs:
            hook_dict["plugs"] = self.plugs

        # Apply passthrough keys.
        hook_dict.update(self.passthrough)
        return hook_dict

    def __repr__(self) -> str:
        return repr(self.__dict__)

    def __str__(self) -> str:
        return str(self.__dict__)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import os
import subprocess
import textwrap
from unittest import mock

import fixtures
import testtools
from testscenarios.scenarios import multiply_scenarios
from testtools.matchers import Equals

from snapcraft import yaml_utils
from snapcraft.internal import errors
from tests import unit
from tests.unit.commands import CommandBaseTestCase


class ScriptletCommandsTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()

        self.make_snapcraft_yaml(
            textwrap.dedent(
                """\
                name: my-snap-name
                base: core18
                summary: summary
                description: description

                adopt-info: my-part
                confinement: devmode

                parts:
                    my-part:
                        plugin: dump
                        source: src
                        override-pull: |
                            snapcraftctl pull
                            snapcraftctl set-grade devel
                            version="$(cat version.txt)"
                            snapcraftctl set-version "$version"
                """
            )
        )

        os.mkdir("src")
        open(os.path.join("src", "version.txt"), "w").write("v1.0")

        fake_install_build_packages = fixtures.MockPatch(
            "snapcraft.internal.lifecycle._runner._install_build_packages",
            return_value=list(),
        )
        self.useFixture(fake_install_build_packages)

        fake_install_build_snaps = fixtures.MockPatch(
            "snapcraft.internal.lifecycle._runner._install_build_snaps",
            return_value=list(),
        )
        self.useFixture(fake_install_build_snaps)

    def test_scriptlet_after_repull(self):
        self.run_command(["prime"])

        with open(os.path.join("prime", "meta", "snap.yaml")) as f:
            y = yaml_utils.load(f)

        self.assertThat(y["grade"], Equals("devel"))
        self.assertThat(y["version"], Equals("v1.0"))

        # modifying source file (src/version.txt) will trigger re-pull
        open(os.path.join("src", "version.txt"), "w").write("v2.0")
        self.run_command(["prime"])

        with open(os.path.join("prime", "meta", "snap.yaml")) as f:
            z = yaml_utils.load(f)

        self.assertThat(z["grade"], Equals("devel"))
        self.assertThat(z["version"], Equals("v2.0"))


class TestScriptletSetter:

    scenarios = [
        (
            "set-version",
            {"setter": "set-version", "getter": "get_version", "value": "v1"},
        ),
        ("set-grade", {"setter": "set-grade", "getter": "get_grade", "value": "devel"}),
    ]

    def test_set_in_pull(self, tmp_work_path, setter, getter, value):
        handler = unit.load_part(
            "test_part",
            part_properties={
                "override-pull": "snapcraftctl {} {}".format(setter, value)
            },
        )

        handler.pull()

        pull_metadata = handler.get_pull_state().scriptlet_metadata

        assert getattr(pull_metadata, getter)() == value

    def test_set_in_build(self, tmp_work_path, setter, getter, value):
        handler = unit.load_part(
            "test_part",
            part_properties={
                "override-build": "snapcraftctl {} {}".format(setter, value)
            },
        )

        handler.pull()
        handler.build()

        pull_metadata = handler.get_pull_state().scriptlet_metadata
        build_metadata = handler.get_build_state().scriptlet_metadata

        assert getattr(pull_metadata, getter)() is None
        assert getattr(build_metadata, getter)() == value

    def test_set_in_stage(self, tmp_work_path, setter, getter, value):
        handler = unit.load_part(
            "test_part",
            part_properties={
                "override-stage": "snapcraftctl {} {}".format(setter, value)
            },
        )

        handler.pull()
        handler.build()
        handler.stage()

        pull_metadata = handler.get_pull_state().scriptlet_metadata
        build_metadata = handler.get_build_state().scriptlet_metadata
        stage_metadata = handler.get_stage_state().scriptlet_metadata

        assert getattr(pull_metadata, getter)() is None
        assert getattr(build_metadata, getter)() is None
        assert getattr(stage_metadata, getter)() == value

    def test_set_in_prime(self, tmp_work_path, setter, getter, value):
        handler = unit.load_part(
            "test_part",
            part_properties={
                "override-prime": "snapcraftctl {} {}".format(setter, value)
            },
        )

        handler.pull()
        handler.build()
        handler.stage()
        handler.prime()

        pull_metadata = handler.get_pull_state().scriptlet_metadata
        build_metadata = handler.get_build_state().scriptlet_metadata
        stage_metadata = handler.get_stage_state().scriptlet_metadata
        prime_metadata = handler.get_prime_state().scriptlet_metadata

        assert getattr(pull_metadata, getter)() is None
        assert getattr(build_metadata, getter)() is None
        assert getattr(stage_metadata, getter)() is None
        assert getattr(prime_metadata, getter)() == value


class TestScriptletMultipleSettersError:

    scriptlet_scenarios = [
        (
            "override-pull/build",
            {
                "override_pull": "snapcraftctl {setter} {value}",
                "override_build": "snapcraftctl {setter} {value}",
                "override_stage": None,
                "override_prime": None,
            },
        ),
        (
            "override-pull/stage",
            {
                "override_pull": "snapcraftctl {setter} {value}",
                "override_build": None,
                "override_stage": "snapcraftctl {setter} {value}",
                "override_prime": None,
            },
        ),
        (
            "override-pull/prime",
            {
                "override_pull": "snapcraftctl {setter} {value}",
                "override_build": None,
                "override_stage": None,
                "override_prime": "snapcraftctl {setter} {value}",
            },
        ),
        (
            "override-build/stage",
            {
                "override_pull": None,
                "override_build": "snapcraftctl {setter} {value}",
                "override_stage": "snapcraftctl {setter} {value}",
                "override_prime": None,
            },
        ),
        (
            "override-build/prime",
            {
                "override_pull": None,
                "override_build": "snapcraftctl {setter} {value}",
                "override_stage": None,
                "override_prime": "snapcraftctl {setter} {value}",
            },
        ),
        (
            "override-stage/prime",
            {
                "override_pull": None,
                "override_build": None,
                "override_stage": "snapcraftctl {setter} {value}",
                "override_prime": "snapcraftctl {setter} {value}",
            },
        ),
    ]

    setter_scenarios = [
        ("set-version", {"setter": "set-version", "value": "v1"}),
        ("set-grade", {"setter": "set-grade", "value": "stable"}),
    ]

    scenarios = multiply_scenarios(setter_scenarios, scriptlet_scenarios)

    def test_set_multiple_times(
        self,
        tmp_work_path,
        setter,
        value,
        override_pull,
        override_build,
        override_stage,
        override_prime,
    ):
        part_properties = {}
        if override_pull is not None:
            part_properties["override-pull"] = override_pull.format(
                setter=setter, value=value
            )
        if override_build is not None:
            part_properties["override-build"] = override_build.format(
                setter=setter, value=value
            )
        if override_stage is not None:
            part_properties["override-stage"] = override_stage.format(
                setter=setter, value=value
            )
        if override_prime is not None:
            part_properties["override-prime"] = override_prime.format(
                setter=setter, value=value
            )

        # A few of these test cases result in only one of these scriptlets
        # being set. In that case, we actually want to double them up (i.e.
        # call set-version twice in the same scriptlet), which should still be
        # an error.
        if len(part_properties) == 1:
            for key, value in part_properties.items():
                part_properties[key] += "\n{}".format(value)

        handler = unit.load_part("test_part", part_properties=part_properties)

        with testtools.ExpectedException(errors.ScriptletDuplicateFieldError):
            silent_popen = functools.partial(
                subprocess.Popen, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )

            with mock.patch("subprocess.Popen", wraps=silent_popen):
                handler.pull()
                handler.build()
                handler.stage()
                handler.prime()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import os
import shutil
from typing import Optional

from snapcraft.file_utils import calculate_hash

from ._cache import SnapcraftCache

logger = logging.getLogger(__name__)


class FileCache(SnapcraftCache):
    """Generic file cache."""

    def __init__(self, *, namespace: str = "files") -> None:
        """Create a FileCache under namespace.

        :param str namespace: set the namespace for the cache
                              (default: "files").
        """
        super().__init__()
        self.file_cache = os.path.join(self.cache_root, namespace)

    def cache(self, *, filename: str, algorithm: str, hash: str) -> Optional[str]:
        """Cache a file revision with hash in XDG cache, unless it already exists.
        :param str filename: path to the file to cache.
        :param str algorithm: algorithm used to calculate the hash as
                              understood by hashlib.
        :param str hash: hash for filename calculated with algorithm.
        :returns: path to cached file.
        """
        # First we verify
        calculated_hash = calculate_hash(filename, algorithm=algorithm)
        if calculated_hash != hash:
            logger.warning(
                "Skipping caching of {!r} as the expected "
                "hash does not match the one "
                "provided".format(filename)
            )
            return None
        cached_file_path = os.path.join(self.file_cache, algorithm, hash)
        os.makedirs(os.path.dirname(cached_file_path), exist_ok=True)
        try:
            if not os.path.isfile(cached_file_path):
                # this must not be hard-linked, as rebuilding a snap
                # with changes should invalidate the cache, hence avoids
                # using fileutils.link_or_copy.
                shutil.copyfile(filename, cached_file_path)
        except OSError:
            logger.warning("Unable to cache file {}.".format(cached_file_path))
            return None
        return cached_file_path

    def get(self, *, algorithm: str, hash: str):
        """Get the filepath which matches the hash calculated with algorithm.

        :param str algorithm: algorithm used to calculate the hash as
                              understood by hashlib.
        :param str hash: hash for filename calculated with algorithm.
        :returns: path to cached file.
        """
        cached_file_path = os.path.join(self.file_cache, algorithm, hash)
        if os.path.exists(cached_file_path):
            logger.debug("Cache hit for hash {!r}".format(hash))
            return cached_file_path
        else:
            return None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from testtools.matchers import Equals

import snapcraft.internal
from snapcraft import yaml_utils
from tests import unit

from .conftest import Project


class PrimeStateBaseTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.project = Project()
        self.files = {"foo"}
        self.directories = {"bar"}
        self.dependency_paths = {"baz"}
        self.part_properties = {
            "override-prime": "touch override-prime",
            "prime": ["qux"],
        }

        self.state = snapcraft.internal.states.PrimeState(
            self.files,
            self.directories,
            self.dependency_paths,
            self.part_properties,
            self.project,
        )


class PrimeStateTestCase(PrimeStateBaseTestCase):
    @mock.patch.object(
        snapcraft.internal.states.PrimeState,
        "__init__",
        wraps=snapcraft.internal.states.PrimeState.__init__,
    )
    def test_yaml_conversion(self, init_spy):
        state_string = yaml_utils.dump(self.state)

        # Verify that the dumped tag was correct
        self.assertThat(state_string.splitlines()[0], Equals("!PrimeState"))

        # Now verify the conversion
        state_from_yaml = yaml_utils.load(state_string)
        self.assertThat(state_from_yaml, Equals(self.state))

        # Verify that init was not called
        init_spy.assert_not_called()

    def test_comparison(self):
        other = snapcraft.internal.states.PrimeState(
            self.files,
            self.directories,
            self.dependency_paths,
            self.part_properties,
            self.project,
        )

        self.assertTrue(self.state == other, "Expected states to be identical")

    def test_properties_of_interest(self):
        properties = self.state.properties_of_interest(self.part_properties)
        self.assertThat(len(properties), Equals(2))
        self.assertThat(properties["override-prime"], Equals("touch override-prime"))
        self.assertThat(properties["prime"], Equals(["qux"]))

    def test_project_options_of_interest(self):
        self.assertFalse(self.state.project_options_of_interest(self.project))


def test_comparison_not_equal(prime_state, prime_state_variant):
    assert prime_state != prime_state_variant
